# -*- coding: utf-8 -*-

#    Copyright (C) 2012 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools
import time

import taskflow.engines
from taskflow.engines.action_engine import executor
from taskflow import exceptions as exc
from taskflow.patterns import graph_flow as gf
from taskflow.patterns import linear_flow as lf
from taskflow.patterns import unordered_flow as uf
from taskflow import retry
from taskflow import states as st
from taskflow import test
from taskflow.tests import utils
from taskflow.types import failure
from taskflow.utils import eventlet_utils as eu

try:
    from taskflow.engines.action_engine import process_executor as pe
except ImportError:
    pe = None


class FailingRetry(retry.Retry):

    def execute(self, **kwargs):
        raise ValueError('OMG I FAILED')

    def revert(self, history, **kwargs):
        self.history = history

    def on_failure(self, **kwargs):
        return retry.REVERT


class NastyFailingRetry(FailingRetry):
    def revert(self, history, **kwargs):
        raise ValueError('WOOT!')


class RetryTest(utils.EngineTestBase):

    def test_run_empty_linear_flow(self):
        flow = lf.Flow('flow-1', utils.OneReturnRetry(provides='x'))
        engine = self._make_engine(flow)
        engine.run()
        self.assertEqual({'x': 1}, engine.storage.fetch_all())

    def test_run_empty_unordered_flow(self):
        flow = uf.Flow('flow-1', utils.OneReturnRetry(provides='x'))
        engine = self._make_engine(flow)
        engine.run()
        self.assertEqual({'x': 1}, engine.storage.fetch_all())

    def test_run_empty_graph_flow(self):
        flow = gf.Flow('flow-1', utils.OneReturnRetry(provides='x'))
        engine = self._make_engine(flow)
        engine.run()
        self.assertEqual({'x': 1}, engine.storage.fetch_all())

    def test_states_retry_success_linear_flow(self):
        flow = lf.Flow('flow-1', retry.Times(4, 'r1', provides='x')).add(
            utils.ProgressingTask("task1"),
            utils.ConditionalTask("task2")
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine) as capturer:
            engine.run()
        self.assertEqual({'y': 2, 'x': 2}, engine.storage.fetch_all())
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING', 'r1.r SUCCESS(1)',
                    'task1.t RUNNING', 'task1.t SUCCESS(5)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task2.t REVERTING', 'task2.t REVERTED(None)',
                    'task1.t REVERTING', 'task1.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task1.t PENDING',
                    'task2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'task2.t RUNNING',
                    'task2.t SUCCESS(None)',
                    'flow-1.f SUCCESS']
        self.assertEqual(expected, capturer.values)

    def test_states_retry_reverted_linear_flow(self):
        flow = lf.Flow('flow-1', retry.Times(2, 'r1', provides='x')).add(
            utils.ProgressingTask("task1"),
            utils.ConditionalTask("task2")
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 4})
        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        self.assertEqual({'y': 4}, engine.storage.fetch_all())
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task1.t PENDING',
                    'task2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

    def test_states_retry_failure_linear_flow(self):
        flow = lf.Flow('flow-1', retry.Times(2, 'r1', provides='x')).add(
            utils.NastyTask("task1"),
            utils.ConditionalTask("task2")
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 4})
        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Gotcha', engine.run)
        self.assertEqual({'y': 4, 'x': 1}, engine.storage.fetch_all())
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(None)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'task1.t REVERTING',
                    'task1.t REVERT_FAILURE(Failure: RuntimeError: Gotcha!)',
                    'flow-1.f FAILURE']
        self.assertEqual(expected, capturer.values)

    def test_states_retry_failure_nested_flow_fails(self):
        flow = lf.Flow('flow-1', utils.retry.AlwaysRevert('r1')).add(
            utils.TaskNoRequiresNoReturns("task1"),
            lf.Flow('flow-2', retry.Times(3, 'r2', provides='x')).add(
                utils.TaskNoRequiresNoReturns("task2"),
                utils.ConditionalTask("task3")
            ),
            utils.TaskNoRequiresNoReturns("task4")
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine) as capturer:
            engine.run()
        self.assertEqual({'y': 2, 'x': 2}, engine.storage.fetch_all())
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(None)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(None)',
                    'r2.r RUNNING',
                    'r2.r SUCCESS(1)',
                    'task2.t RUNNING',
                    'task2.t SUCCESS(None)',
                    'task3.t RUNNING',
                    'task3.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task3.t REVERTING',
                    'task3.t REVERTED(None)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r2.r RETRYING',
                    'task2.t PENDING',
                    'task3.t PENDING',
                    'r2.r RUNNING',
                    'r2.r SUCCESS(2)',
                    'task2.t RUNNING',
                    'task2.t SUCCESS(None)',
                    'task3.t RUNNING',
                    'task3.t SUCCESS(None)',
                    'task4.t RUNNING',
                    'task4.t SUCCESS(None)',
                    'flow-1.f SUCCESS']
        self.assertEqual(expected, capturer.values)

    def test_new_revert_vs_old(self):
        flow = lf.Flow('flow-1').add(
            utils.TaskNoRequiresNoReturns("task1"),
            lf.Flow('flow-2', retry.Times(1, 'r1', provides='x')).add(
                utils.TaskNoRequiresNoReturns("task2"),
                utils.ConditionalTask("task3")
            ),
            utils.TaskNoRequiresNoReturns("task4")
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine) as capturer:
            try:
                engine.run()
            except Exception:
                pass

        expected = ['flow-1.f RUNNING',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(None)',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    'task2.t RUNNING',
                    'task2.t SUCCESS(None)',
                    'task3.t RUNNING',
                    'task3.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task3.t REVERTING',
                    'task3.t REVERTED(None)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

        engine = self._make_engine(flow, defer_reverts=True)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine) as capturer:
            try:
                engine.run()
            except Exception:
                pass

        expected = ['flow-1.f RUNNING',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(None)',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    'task2.t RUNNING',
                    'task2.t SUCCESS(None)',
                    'task3.t RUNNING',
                    'task3.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task3.t REVERTING',
                    'task3.t REVERTED(None)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

    def test_states_retry_failure_parent_flow_fails(self):
        flow = lf.Flow('flow-1', retry.Times(3, 'r1', provides='x1')).add(
            utils.TaskNoRequiresNoReturns("task1"),
            lf.Flow('flow-2', retry.Times(3, 'r2', provides='x2')).add(
                utils.TaskNoRequiresNoReturns("task2"),
                utils.TaskNoRequiresNoReturns("task3")
            ),
            utils.ConditionalTask("task4", rebind={'x': 'x1'})
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine) as capturer:
            engine.run()
        self.assertEqual({'y': 2, 'x1': 2,
                          'x2': 1},
                         engine.storage.fetch_all())
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(None)',
                    'r2.r RUNNING',
                    'r2.r SUCCESS(1)',
                    'task2.t RUNNING',
                    'task2.t SUCCESS(None)',
                    'task3.t RUNNING',
                    'task3.t SUCCESS(None)',
                    'task4.t RUNNING',
                    'task4.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task4.t REVERTING',
                    'task4.t REVERTED(None)',
                    'task3.t REVERTING',
                    'task3.t REVERTED(None)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r2.r REVERTING',
                    'r2.r REVERTED(None)',
                    'task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task1.t PENDING',
                    'r2.r PENDING',
                    'task2.t PENDING',
                    'task3.t PENDING',
                    'task4.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(None)',
                    'r2.r RUNNING',
                    'r2.r SUCCESS(1)',
                    'task2.t RUNNING',
                    'task2.t SUCCESS(None)',
                    'task3.t RUNNING',
                    'task3.t SUCCESS(None)',
                    'task4.t RUNNING',
                    'task4.t SUCCESS(None)',
                    'flow-1.f SUCCESS']
        self.assertEqual(expected, capturer.values)

    def test_unordered_flow_task_fails_parallel_tasks_should_be_reverted(self):
        flow = uf.Flow('flow-1', retry.Times(3, 'r', provides='x')).add(
            utils.ProgressingTask("task1"),
            utils.ConditionalTask("task2")
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine) as capturer:
            engine.run()
        self.assertEqual({'y': 2, 'x': 2}, engine.storage.fetch_all())
        expected = ['flow-1.f RUNNING',
                    'r.r RUNNING',
                    'r.r SUCCESS(1)',
                    'task1.t RUNNING',
                    'task2.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'task2.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task2.t REVERTING',
                    'task1.t REVERTING',
                    'task2.t REVERTED(None)',
                    'task1.t REVERTED(None)',
                    'r.r RETRYING',
                    'task1.t PENDING',
                    'task2.t PENDING',
                    'r.r RUNNING',
                    'r.r SUCCESS(2)',
                    'task1.t RUNNING',
                    'task2.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'task2.t SUCCESS(None)',
                    'flow-1.f SUCCESS']
        self.assertCountEqual(capturer.values, expected)

    def test_nested_flow_reverts_parent_retries(self):
        retry1 = retry.Times(3, 'r1', provides='x')
        retry2 = retry.Times(0, 'r2', provides='x2')
        flow = lf.Flow('flow-1', retry1).add(
            utils.ProgressingTask("task1"),
            lf.Flow('flow-2', retry2).add(utils.ConditionalTask("task2"))
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine) as capturer:
            engine.run()
        self.assertEqual({'y': 2, 'x': 2, 'x2': 1}, engine.storage.fetch_all())
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'r2.r RUNNING',
                    'r2.r SUCCESS(1)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r2.r REVERTING',
                    'r2.r REVERTED(None)',
                    'task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task1.t PENDING',
                    'r2.r PENDING',
                    'task2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'r2.r RUNNING',
                    'r2.r SUCCESS(1)',
                    'task2.t RUNNING',
                    'task2.t SUCCESS(None)',
                    'flow-1.f SUCCESS']
        self.assertEqual(expected, capturer.values)

    def test_nested_flow_with_retry_revert(self):
        retry1 = retry.Times(0, 'r1', provides='x2')
        flow = lf.Flow('flow-1').add(
            utils.ProgressingTask("task1"),
            lf.Flow('flow-2', retry1).add(
                utils.ConditionalTask("task2", inject={'x': 1}))
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine) as capturer:
            try:
                engine.run()
            except Exception:
                pass
        self.assertEqual({'y': 2}, engine.storage.fetch_all())
        expected = ['flow-1.f RUNNING',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

    def test_nested_flow_with_retry_revert_all(self):
        retry1 = retry.Times(0, 'r1', provides='x2', revert_all=True)
        flow = lf.Flow('flow-1').add(
            utils.ProgressingTask("task1"),
            lf.Flow('flow-2', retry1).add(
                utils.ConditionalTask("task2", inject={'x': 1}))
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine) as capturer:
            try:
                engine.run()
            except Exception:
                pass
        self.assertEqual({'y': 2}, engine.storage.fetch_all())
        expected = ['flow-1.f RUNNING',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

    def test_revert_all_retry(self):
        flow = lf.Flow('flow-1', retry.Times(3, 'r1', provides='x')).add(
            utils.ProgressingTask("task1"),
            lf.Flow('flow-2', retry.AlwaysRevertAll('r2')).add(
                utils.ConditionalTask("task2"))
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        self.assertEqual({'y': 2}, engine.storage.fetch_all())
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'r2.r RUNNING',
                    'r2.r SUCCESS(None)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r2.r REVERTING',
                    'r2.r REVERTED(None)',
                    'task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

    def test_restart_reverted_flow_with_retry(self):
        flow = lf.Flow('test', retry=utils.OneReturnRetry(provides='x')).add(
            utils.FailingTask('fail'))
        engine = self._make_engine(flow)
        self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)

    def test_restart_reverted_unordered_flows_with_retries(self):
        now = time.time()

        # First flow of an unordered flow:
        subflow1 = lf.Flow('subflow1')

        # * a task that completes in 3 sec with a few retries
        subsubflow1 = lf.Flow('subflow1.subsubflow1',
                              retry=utils.RetryFiveTimes())
        subsubflow1.add(utils.SuccessAfter3Sec('subflow1.fail1',
                                               inject={'start_time': now}))
        subflow1.add(subsubflow1)

        # * a task that fails and triggers a revert after 5 retries
        subsubflow2 = lf.Flow('subflow1.subsubflow2',
                              retry=utils.RetryFiveTimes())
        subsubflow2.add(utils.FailingTask('subflow1.fail2'))
        subflow1.add(subsubflow2)

        # Second flow of the unordered flow:
        subflow2 = lf.Flow('subflow2')

        # * a task that always fails and retries
        subsubflow1 = lf.Flow('subflow2.subsubflow1',
                              retry=utils.AlwaysRetry())
        subsubflow1.add(utils.FailingTask('subflow2.fail1'))
        subflow2.add(subsubflow1)

        unordered_flow = uf.Flow('unordered_flow')
        unordered_flow.add(subflow1, subflow2)

        # Main flow, contains a simple task and an unordered flow
        flow = lf.Flow('test')
        flow.add(utils.NoopTask('task1'))
        flow.add(unordered_flow)

        engine = self._make_engine(flow)

        # This test fails when using Green threads, skipping it for now
        if isinstance(engine._task_executor,
                      executor.ParallelGreenThreadTaskExecutor):
            self.skipTest("Skipping this test when using green threads.")

        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(exc.WrappedFailure,
                                   '.*RuntimeError: Woot!',
                                   engine.run)
            # task1 should have been reverted
            self.assertIn('task1.t REVERTED(None)', capturer.values)

    def test_run_just_retry(self):
        flow = utils.OneReturnRetry(provides='x')
        engine = self._make_engine(flow)
        self.assertRaises(TypeError, engine.run)

    def test_use_retry_as_a_task(self):
        flow = lf.Flow('test').add(utils.OneReturnRetry(provides='x'))
        engine = self._make_engine(flow)
        self.assertRaises(TypeError, engine.run)

    def test_resume_flow_that_had_been_interrupted_during_retrying(self):
        flow = lf.Flow('flow-1', retry.Times(3, 'r1')).add(
            utils.ProgressingTask('t1'),
            utils.ProgressingTask('t2'),
            utils.ProgressingTask('t3')
        )
        engine = self._make_engine(flow)
        engine.compile()
        engine.prepare()
        with utils.CaptureListener(engine) as capturer:
            engine.storage.set_atom_state('r1', st.RETRYING)
            engine.storage.set_atom_state('t1', st.PENDING)
            engine.storage.set_atom_state('t2', st.REVERTED)
            engine.storage.set_atom_state('t3', st.REVERTED)
            engine.run()
        expected = ['flow-1.f RUNNING',
                    't2.t PENDING',
                    't3.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    't1.t RUNNING',
                    't1.t SUCCESS(5)',
                    't2.t RUNNING',
                    't2.t SUCCESS(5)',
                    't3.t RUNNING',
                    't3.t SUCCESS(5)',
                    'flow-1.f SUCCESS']
        self.assertEqual(expected, capturer.values)

    def test_resume_flow_that_should_be_retried(self):
        flow = lf.Flow('flow-1', retry.Times(3, 'r1')).add(
            utils.ProgressingTask('t1'),
            utils.ProgressingTask('t2')
        )
        engine = self._make_engine(flow)
        engine.compile()
        engine.prepare()
        with utils.CaptureListener(engine) as capturer:
            engine.storage.set_atom_intention('r1', st.RETRY)
            engine.storage.set_atom_state('r1', st.SUCCESS)
            engine.storage.set_atom_state('t1', st.REVERTED)
            engine.storage.set_atom_state('t2', st.REVERTED)
            engine.run()
        expected = ['flow-1.f RUNNING',
                    'r1.r RETRYING',
                    't1.t PENDING',
                    't2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    't1.t RUNNING',
                    't1.t SUCCESS(5)',
                    't2.t RUNNING',
                    't2.t SUCCESS(5)',
                    'flow-1.f SUCCESS']
        self.assertEqual(expected, capturer.values)

    def test_retry_tasks_that_has_not_been_reverted(self):
        flow = lf.Flow('flow-1', retry.Times(3, 'r1', provides='x')).add(
            utils.ConditionalTask('c'),
            utils.ProgressingTask('t1')
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine) as capturer:
            engine.run()
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    'c.t RUNNING',
                    'c.t FAILURE(Failure: RuntimeError: Woot!)',
                    'c.t REVERTING',
                    'c.t REVERTED(None)',
                    'r1.r RETRYING',
                    'c.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    'c.t RUNNING',
                    'c.t SUCCESS(None)',
                    't1.t RUNNING',
                    't1.t SUCCESS(5)',
                    'flow-1.f SUCCESS']
        self.assertEqual(expected, capturer.values)

    def test_default_times_retry(self):
        flow = lf.Flow('flow-1', retry.Times(3, 'r1')).add(
            utils.ProgressingTask('t1'),
            utils.FailingTask('t2'))
        engine = self._make_engine(flow)
        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(1)',
                    't1.t RUNNING',
                    't1.t SUCCESS(5)',
                    't2.t RUNNING',
                    't2.t FAILURE(Failure: RuntimeError: Woot!)',
                    't2.t REVERTING',
                    't2.t REVERTED(None)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r RETRYING',
                    't1.t PENDING',
                    't2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    't1.t RUNNING',
                    't1.t SUCCESS(5)',
                    't2.t RUNNING',
                    't2.t FAILURE(Failure: RuntimeError: Woot!)',
                    't2.t REVERTING',
                    't2.t REVERTED(None)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r RETRYING',
                    't1.t PENDING',
                    't2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(3)',
                    't1.t RUNNING',
                    't1.t SUCCESS(5)',
                    't2.t RUNNING',
                    't2.t FAILURE(Failure: RuntimeError: Woot!)',
                    't2.t REVERTING',
                    't2.t REVERTED(None)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

    def test_for_each_with_list(self):
        collection = [3, 2, 3, 5]
        retry1 = retry.ForEach(collection, 'r1', provides='x')
        flow = lf.Flow('flow-1', retry1).add(utils.FailingTaskWithOneArg('t1'))
        engine = self._make_engine(flow)
        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(3)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 3)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r RETRYING',
                    't1.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 2)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r RETRYING',
                    't1.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(3)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 3)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r RETRYING',
                    't1.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(5)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 5)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

    def test_for_each_with_set(self):
        collection = set([3, 2, 5])
        retry1 = retry.ForEach(collection, 'r1', provides='x')
        flow = lf.Flow('flow-1', retry1).add(utils.FailingTaskWithOneArg('t1'))
        engine = self._make_engine(flow)
        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 2)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r RETRYING',
                    't1.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(3)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 3)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r RETRYING',
                    't1.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(5)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 5)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertCountEqual(capturer.values, expected)

    def test_nested_for_each_revert(self):
        collection = [3, 2, 3, 5]
        retry1 = retry.ForEach(collection, 'r1', provides='x')
        flow = lf.Flow('flow-1').add(
            utils.ProgressingTask("task1"),
            lf.Flow('flow-2', retry1).add(
                utils.FailingTaskWithOneArg('task2')
            )
        )
        engine = self._make_engine(flow)
        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        expected = ['flow-1.f RUNNING',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(3)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot with 3)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot with 2)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(3)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot with 3)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(5)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot with 5)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

    def test_nested_for_each_revert_all(self):
        collection = [3, 2, 3, 5]
        retry1 = retry.ForEach(collection, 'r1', provides='x', revert_all=True)
        flow = lf.Flow('flow-1').add(
            utils.ProgressingTask("task1"),
            lf.Flow('flow-2', retry1).add(
                utils.FailingTaskWithOneArg('task2')
            )
        )
        engine = self._make_engine(flow)
        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        expected = ['flow-1.f RUNNING',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(3)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot with 3)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot with 2)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(3)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot with 3)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(5)',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot with 5)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

    def test_for_each_empty_collection(self):
        values = []
        retry1 = retry.ForEach(values, 'r1', provides='x')
        flow = lf.Flow('flow-1', retry1).add(utils.ConditionalTask('t1'))
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 1})
        self.assertRaisesRegex(exc.NotFound, '^No elements left', engine.run)

    def test_parameterized_for_each_with_list(self):
        values = [3, 2, 5]
        retry1 = retry.ParameterizedForEach('r1', provides='x')
        flow = lf.Flow('flow-1', retry1).add(utils.FailingTaskWithOneArg('t1'))
        engine = self._make_engine(flow)
        engine.storage.inject({'values': values, 'y': 1})
        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(3)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 3)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r RETRYING',
                    't1.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 2)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r RETRYING',
                    't1.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(5)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 5)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

    def test_parameterized_for_each_with_set(self):
        values = ([3, 2, 5])
        retry1 = retry.ParameterizedForEach('r1', provides='x')
        flow = lf.Flow('flow-1', retry1).add(utils.FailingTaskWithOneArg('t1'))
        engine = self._make_engine(flow)
        engine.storage.inject({'values': values, 'y': 1})
        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        expected = ['flow-1.f RUNNING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(3)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 3)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r RETRYING',
                    't1.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 2)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r RETRYING',
                    't1.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(5)',
                    't1.t RUNNING',
                    't1.t FAILURE(Failure: RuntimeError: Woot with 5)',
                    't1.t REVERTING',
                    't1.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertCountEqual(capturer.values, expected)

    def test_nested_parameterized_for_each_revert(self):
        values = [3, 2, 5]
        retry1 = retry.ParameterizedForEach('r1', provides='x')
        flow = lf.Flow('flow-1').add(
            utils.ProgressingTask('task-1'),
            lf.Flow('flow-2', retry1).add(
                utils.FailingTaskWithOneArg('task-2')
            )
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'values': values, 'y': 1})
        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        expected = ['flow-1.f RUNNING',
                    'task-1.t RUNNING',
                    'task-1.t SUCCESS(5)',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(3)',
                    'task-2.t RUNNING',
                    'task-2.t FAILURE(Failure: RuntimeError: Woot with 3)',
                    'task-2.t REVERTING',
                    'task-2.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task-2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    'task-2.t RUNNING',
                    'task-2.t FAILURE(Failure: RuntimeError: Woot with 2)',
                    'task-2.t REVERTING',
                    'task-2.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task-2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(5)',
                    'task-2.t RUNNING',
                    'task-2.t FAILURE(Failure: RuntimeError: Woot with 5)',
                    'task-2.t REVERTING',
                    'task-2.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

    def test_nested_parameterized_for_each_revert_all(self):
        values = [3, 2, 5]
        retry1 = retry.ParameterizedForEach('r1', provides='x',
                                            revert_all=True)
        flow = lf.Flow('flow-1').add(
            utils.ProgressingTask('task-1'),
            lf.Flow('flow-2', retry1).add(
                utils.FailingTaskWithOneArg('task-2')
            )
        )
        engine = self._make_engine(flow)
        engine.storage.inject({'values': values, 'y': 1})
        with utils.CaptureListener(engine) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        expected = ['flow-1.f RUNNING',
                    'task-1.t RUNNING',
                    'task-1.t SUCCESS(5)',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(3)',
                    'task-2.t RUNNING',
                    'task-2.t FAILURE(Failure: RuntimeError: Woot with 3)',
                    'task-2.t REVERTING',
                    'task-2.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task-2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(2)',
                    'task-2.t RUNNING',
                    'task-2.t FAILURE(Failure: RuntimeError: Woot with 2)',
                    'task-2.t REVERTING',
                    'task-2.t REVERTED(None)',
                    'r1.r RETRYING',
                    'task-2.t PENDING',
                    'r1.r RUNNING',
                    'r1.r SUCCESS(5)',
                    'task-2.t RUNNING',
                    'task-2.t FAILURE(Failure: RuntimeError: Woot with 5)',
                    'task-2.t REVERTING',
                    'task-2.t REVERTED(None)',
                    'r1.r REVERTING',
                    'r1.r REVERTED(None)',
                    'task-1.t REVERTING',
                    'task-1.t REVERTED(None)',
                    'flow-1.f REVERTED']
        self.assertEqual(expected, capturer.values)

    def test_parameterized_for_each_empty_collection(self):
        values = []
        retry1 = retry.ParameterizedForEach('r1', provides='x')
        flow = lf.Flow('flow-1', retry1).add(utils.ConditionalTask('t1'))
        engine = self._make_engine(flow)
        engine.storage.inject({'values': values, 'y': 1})
        self.assertRaisesRegex(exc.NotFound, '^No elements left', engine.run)

    def _pretend_to_run_a_flow_and_crash(self, when):
        flow = uf.Flow('flow-1', retry.Times(3, provides='x')).add(
            utils.ProgressingTask('task1'))
        engine = self._make_engine(flow)
        engine.compile()
        engine.prepare()
        # imagine we run engine
        engine.storage.set_flow_state(st.RUNNING)
        engine.storage.set_atom_intention('flow-1_retry', st.EXECUTE)
        engine.storage.set_atom_intention('task1', st.EXECUTE)
        # we execute retry
        engine.storage.save('flow-1_retry', 1)
        # task fails
        fail = failure.Failure.from_exception(RuntimeError('foo'))
        engine.storage.save('task1', fail, state=st.FAILURE)
        if when == 'task fails':
            return engine
        # we save it's failure to retry and ask what to do
        engine.storage.save_retry_failure('flow-1_retry', 'task1', fail)
        if when == 'retry queried':
            return engine
        # it returned 'RETRY', so we update it's intention
        engine.storage.set_atom_intention('flow-1_retry', st.RETRY)
        if when == 'retry updated':
            return engine
        # we set task1 intention to REVERT
        engine.storage.set_atom_intention('task1', st.REVERT)
        if when == 'task updated':
            return engine
        # we schedule task1 for reversion
        engine.storage.set_atom_state('task1', st.REVERTING)
        if when == 'revert scheduled':
            return engine
        raise ValueError('Invalid crash point: %s' % when)

    def test_resumption_on_crash_after_task_failure(self):
        engine = self._pretend_to_run_a_flow_and_crash('task fails')
        with utils.CaptureListener(engine) as capturer:
            engine.run()
        expected = ['task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'flow-1_retry.r RETRYING',
                    'task1.t PENDING',
                    'flow-1_retry.r RUNNING',
                    'flow-1_retry.r SUCCESS(2)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'flow-1.f SUCCESS']
        self.assertEqual(expected, capturer.values)

    def test_resumption_on_crash_after_retry_queried(self):
        engine = self._pretend_to_run_a_flow_and_crash('retry queried')
        with utils.CaptureListener(engine) as capturer:
            engine.run()
        expected = ['task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'flow-1_retry.r RETRYING',
                    'task1.t PENDING',
                    'flow-1_retry.r RUNNING',
                    'flow-1_retry.r SUCCESS(2)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'flow-1.f SUCCESS']
        self.assertEqual(expected, capturer.values)

    def test_resumption_on_crash_after_retry_updated(self):
        engine = self._pretend_to_run_a_flow_and_crash('retry updated')
        with utils.CaptureListener(engine) as capturer:
            engine.run()
        expected = ['task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'flow-1_retry.r RETRYING',
                    'task1.t PENDING',
                    'flow-1_retry.r RUNNING',
                    'flow-1_retry.r SUCCESS(2)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'flow-1.f SUCCESS']
        self.assertEqual(expected, capturer.values)

    def test_resumption_on_crash_after_task_updated(self):
        engine = self._pretend_to_run_a_flow_and_crash('task updated')
        with utils.CaptureListener(engine) as capturer:
            engine.run()
        expected = ['task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'flow-1_retry.r RETRYING',
                    'task1.t PENDING',
                    'flow-1_retry.r RUNNING',
                    'flow-1_retry.r SUCCESS(2)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'flow-1.f SUCCESS']
        self.assertEqual(expected, capturer.values)

    def test_resumption_on_crash_after_revert_scheduled(self):
        engine = self._pretend_to_run_a_flow_and_crash('revert scheduled')
        with utils.CaptureListener(engine) as capturer:
            engine.run()
        expected = ['task1.t REVERTED(None)',
                    'flow-1_retry.r RETRYING',
                    'task1.t PENDING',
                    'flow-1_retry.r RUNNING',
                    'flow-1_retry.r SUCCESS(2)',
                    'task1.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'flow-1.f SUCCESS']
        self.assertEqual(expected, capturer.values)

    def test_retry_fails(self):
        r = FailingRetry()
        flow = lf.Flow('testflow', r)
        engine = self._make_engine(flow)
        self.assertRaisesRegex(ValueError, '^OMG', engine.run)
        self.assertEqual(1, len(engine.storage.get_retry_histories()))
        self.assertEqual(0, len(r.history))
        self.assertEqual([], list(r.history.outcomes_iter()))
        self.assertIsNotNone(r.history.failure)
        self.assertTrue(r.history.caused_by(ValueError, include_retry=True))

    def test_retry_revert_fails(self):
        r = NastyFailingRetry()
        flow = lf.Flow('testflow', r)
        engine = self._make_engine(flow)
        self.assertRaisesRegex(ValueError, '^WOOT', engine.run)

    def test_nested_provides_graph_reverts_correctly(self):
        flow = gf.Flow("test").add(
            utils.ProgressingTask('a', requires=['x']),
            lf.Flow("test2", retry=retry.Times(2)).add(
                utils.ProgressingTask('b', provides='x'),
                utils.FailingTask('c')))
        engine = self._make_engine(flow)
        engine.compile()
        engine.prepare()
        engine.storage.save('test2_retry', 1)
        engine.storage.save('b', 11)
        engine.storage.save('a', 10)
        with utils.CaptureListener(engine, capture_flow=False) as capturer:
            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
        expected = ['c.t RUNNING',
                    'c.t FAILURE(Failure: RuntimeError: Woot!)',
                    'a.t REVERTING',
                    'c.t REVERTING',
                    'a.t REVERTED(None)',
                    'c.t REVERTED(None)',
                    'b.t REVERTING',
                    'b.t REVERTED(None)']
        self.assertCountEqual(capturer.values[:8], expected)
        # Task 'a' was or was not executed again, both cases are ok.
        self.assertIsSuperAndSubsequence(capturer.values[8:], [
            'b.t RUNNING',
            'c.t FAILURE(Failure: RuntimeError: Woot!)',
            'b.t REVERTED(None)',
        ])
        self.assertEqual(st.REVERTED, engine.storage.get_flow_state())

    def test_nested_provides_graph_retried_correctly(self):
        flow = gf.Flow("test").add(
            utils.ProgressingTask('a', requires=['x']),
            lf.Flow("test2", retry=retry.Times(2)).add(
                utils.ProgressingTask('b', provides='x'),
                utils.ProgressingTask('c')))
        engine = self._make_engine(flow)
        engine.compile()
        engine.prepare()
        engine.storage.save('test2_retry', 1)
        engine.storage.save('b', 11)
        # pretend that 'c' failed
        fail = failure.Failure.from_exception(RuntimeError('Woot!'))
        engine.storage.save('c', fail, st.FAILURE)
        with utils.CaptureListener(engine, capture_flow=False) as capturer:
            engine.run()
        expected = ['c.t REVERTING',
                    'c.t REVERTED(None)',
                    'b.t REVERTING',
                    'b.t REVERTED(None)']
        self.assertCountEqual(capturer.values[:4], expected)
        expected = ['test2_retry.r RETRYING',
                    'b.t PENDING',
                    'c.t PENDING',
                    'test2_retry.r RUNNING',
                    'test2_retry.r SUCCESS(2)',
                    'b.t RUNNING',
                    'b.t SUCCESS(5)',
                    'a.t RUNNING',
                    'c.t RUNNING',
                    'a.t SUCCESS(5)',
                    'c.t SUCCESS(5)']
        self.assertCountEqual(expected, capturer.values[4:])
        self.assertEqual(st.SUCCESS, engine.storage.get_flow_state())


class RetryParallelExecutionTest(utils.EngineTestBase):
    # FIXME(harlowja): fix this class so that it doesn't use events or uses
    # them in a way that works with more executors...

    def test_when_subflow_fails_revert_running_tasks(self):
        waiting_task = utils.WaitForOneFromTask('task1', 'task2',
                                                [st.SUCCESS, st.FAILURE])
        flow = uf.Flow('flow-1', retry.Times(3, 'r', provides='x')).add(
            waiting_task,
            utils.ConditionalTask('task2')
        )
        engine = self._make_engine(flow)
        engine.atom_notifier.register('*', waiting_task.callback)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine, capture_flow=False) as capturer:
            engine.run()
        self.assertEqual({'y': 2, 'x': 2}, engine.storage.fetch_all())
        expected = ['r.r RUNNING',
                    'r.r SUCCESS(1)',
                    'task1.t RUNNING',
                    'task2.t RUNNING',
                    'task2.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'task1.t SUCCESS(5)',
                    'task1.t REVERTING',
                    'task1.t REVERTED(None)',
                    'r.r RETRYING',
                    'task1.t PENDING',
                    'task2.t PENDING',
                    'r.r RUNNING',
                    'r.r SUCCESS(2)',
                    'task1.t RUNNING',
                    'task2.t RUNNING',
                    'task2.t SUCCESS(None)',
                    'task1.t SUCCESS(5)']
        self.assertCountEqual(capturer.values, expected)

    def test_when_subflow_fails_revert_success_tasks(self):
        waiting_task = utils.WaitForOneFromTask('task2', 'task1',
                                                [st.SUCCESS, st.FAILURE])
        flow = uf.Flow('flow-1', retry.Times(3, 'r', provides='x')).add(
            utils.ProgressingTask('task1'),
            lf.Flow('flow-2').add(
                waiting_task,
                utils.ConditionalTask('task3'))
        )
        engine = self._make_engine(flow)
        engine.atom_notifier.register('*', waiting_task.callback)
        engine.storage.inject({'y': 2})
        with utils.CaptureListener(engine, capture_flow=False) as capturer:
            engine.run()
        self.assertEqual({'y': 2, 'x': 2}, engine.storage.fetch_all())
        expected = ['r.r RUNNING',
                    'r.r SUCCESS(1)',
                    'task1.t RUNNING',
                    'task2.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'task2.t SUCCESS(5)',
                    'task3.t RUNNING',
                    'task3.t FAILURE(Failure: RuntimeError: Woot!)',
                    'task3.t REVERTING',
                    'task1.t REVERTING',
                    'task3.t REVERTED(None)',
                    'task1.t REVERTED(None)',
                    'task2.t REVERTING',
                    'task2.t REVERTED(None)',
                    'r.r RETRYING',
                    'task1.t PENDING',
                    'task2.t PENDING',
                    'task3.t PENDING',
                    'r.r RUNNING',
                    'r.r SUCCESS(2)',
                    'task1.t RUNNING',
                    'task2.t RUNNING',
                    'task1.t SUCCESS(5)',
                    'task2.t SUCCESS(5)',
                    'task3.t RUNNING',
                    'task3.t SUCCESS(None)']
        self.assertCountEqual(capturer.values, expected)


class SerialEngineTest(RetryTest, test.TestCase):
    def _make_engine(self, flow, defer_reverts=None, flow_detail=None):
        return taskflow.engines.load(flow,
                                     flow_detail=flow_detail,
                                     engine='serial',
                                     backend=self.backend,
                                     defer_reverts=defer_reverts)


class ParallelEngineWithThreadsTest(RetryTest,
                                    RetryParallelExecutionTest,
                                    test.TestCase):
    _EXECUTOR_WORKERS = 2

    def _make_engine(self, flow, defer_reverts=None, flow_detail=None,
                     executor=None):
        if executor is None:
            executor = 'threads'
        return taskflow.engines.load(flow,
                                     flow_detail=flow_detail,
                                     engine='parallel',
                                     backend=self.backend,
                                     executor=executor,
                                     max_workers=self._EXECUTOR_WORKERS,
                                     defer_reverts=defer_reverts)


@testtools.skipIf(not eu.EVENTLET_AVAILABLE, 'eventlet is not available')
class ParallelEngineWithEventletTest(RetryTest, test.TestCase):

    def _make_engine(self, flow, defer_reverts=None, flow_detail=None,
                     executor=None):
        if executor is None:
            executor = 'greenthreads'
        return taskflow.engines.load(flow,
                                     flow_detail=flow_detail,
                                     backend=self.backend,
                                     engine='parallel',
                                     executor=executor,
                                     defer_reverts=defer_reverts)


@testtools.skipIf(pe is None, 'process_executor is not available')
class ParallelEngineWithProcessTest(RetryTest, test.TestCase):
    _EXECUTOR_WORKERS = 2

    def _make_engine(self, flow, defer_reverts=None, flow_detail=None,
                     executor=None):
        if executor is None:
            executor = 'processes'
        return taskflow.engines.load(flow,
                                     flow_detail=flow_detail,
                                     engine='parallel',
                                     backend=self.backend,
                                     executor=executor,
                                     max_workers=self._EXECUTOR_WORKERS,
                                     defer_reverts=defer_reverts)



# -*- coding: utf-8 -*-

#    Copyright (C) 2012 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import io
import os

import networkx as nx
from networkx.drawing import nx_pydot


def _common_format(g, edge_notation):
    lines = []
    lines.append("Name: %s" % g.name)
    lines.append("Type: %s" % type(g).__name__)
    lines.append("Frozen: %s" % nx.is_frozen(g))
    lines.append("Density: %0.3f" % nx.density(g))
    lines.append("Nodes: %s" % g.number_of_nodes())
    for n, n_data in g.nodes(data=True):
        if n_data:
            lines.append("  - %s (%s)" % (n, n_data))
        else:
            lines.append("  - %s" % n)
    lines.append("Edges: %s" % g.number_of_edges())
    for (u, v, e_data) in g.edges(data=True):
        if e_data:
            lines.append("  %s %s %s (%s)" % (u, edge_notation, v, e_data))
        else:
            lines.append("  %s %s %s" % (u, edge_notation, v))
    return lines


class Graph(nx.Graph):
    """A graph subclass with useful utility functions."""

    def __init__(self, incoming_graph_data=None, name=''):
        super(Graph, self).__init__(incoming_graph_data=incoming_graph_data,
                                    name=name)
        self.frozen = False

    def freeze(self):
        """Freezes the graph so that no more mutations can occur."""
        if not self.frozen:
            nx.freeze(self)
        return self

    def export_to_dot(self):
        """Exports the graph to a dot format (requires pydot library)."""
        return nx_pydot.to_pydot(self).to_string()

    def pformat(self):
        """Pretty formats your graph into a string."""
        return os.linesep.join(_common_format(self, "<->"))

    def add_edge(self, u, v, attr_dict=None, **attr):
        """Add an edge between u and v."""
        if attr_dict is not None:
            return super(Graph, self).add_edge(u, v, **attr_dict)
        return super(Graph, self).add_edge(u, v, **attr)

    def add_node(self, n, attr_dict=None, **attr):
        """Add a single node n and update node attributes."""
        if attr_dict is not None:
            return super(Graph, self).add_node(n, **attr_dict)
        return super(Graph, self).add_node(n, **attr)

    def fresh_copy(self):
        """Return a fresh copy graph with the same data structure.

        A fresh copy has no nodes, edges or graph attributes. It is
        the same data structure as the current graph. This method is
        typically used to create an empty version of the graph.
        """
        return Graph()


class DiGraph(nx.DiGraph):
    """A directed graph subclass with useful utility functions."""

    def __init__(self, incoming_graph_data=None, name=''):
        super(DiGraph, self).__init__(incoming_graph_data=incoming_graph_data,
                                      name=name)
        self.frozen = False

    def freeze(self):
        """Freezes the graph so that no more mutations can occur."""
        if not self.frozen:
            nx.freeze(self)
        return self

    def get_edge_data(self, u, v, default=None):
        """Returns a *copy* of the edge attribute dictionary between (u, v).

        NOTE(harlowja): this differs from the networkx get_edge_data() as that
        function does not return a copy (but returns a reference to the actual
        edge data).
        """
        try:
            return dict(self.adj[u][v])
        except KeyError:
            return default

    def topological_sort(self):
        """Return a list of nodes in this graph in topological sort order."""
        return nx.topological_sort(self)

    def pformat(self):
        """Pretty formats your graph into a string.

        This pretty formatted string representation includes many useful
        details about your graph, including; name, type, frozeness, node count,
        nodes, edge count, edges, graph density and graph cycles (if any).
        """
        lines = _common_format(self, "->")
        cycles = list(nx.cycles.recursive_simple_cycles(self))
        lines.append("Cycles: %s" % len(cycles))
        for cycle in cycles:
            buf = io.StringIO()
            buf.write("%s" % (cycle[0]))
            for i in range(1, len(cycle)):
                buf.write(" --> %s" % (cycle[i]))
            buf.write(" --> %s" % (cycle[0]))
            lines.append("  %s" % buf.getvalue())
        return os.linesep.join(lines)

    def export_to_dot(self):
        """Exports the graph to a dot format (requires pydot library)."""
        return nx_pydot.to_pydot(self).to_string()

    def is_directed_acyclic(self):
        """Returns if this graph is a DAG or not."""
        return nx.is_directed_acyclic_graph(self)

    def no_successors_iter(self):
        """Returns an iterator for all nodes with no successors."""
        for n in self.nodes:
            if not len(list(self.successors(n))):
                yield n

    def no_predecessors_iter(self):
        """Returns an iterator for all nodes with no predecessors."""
        for n in self.nodes:
            if not len(list(self.predecessors(n))):
                yield n

    def bfs_predecessors_iter(self, n):
        """Iterates breadth first over *all* predecessors of a given node.

        This will go through the nodes predecessors, then the predecessor nodes
        predecessors and so on until no more predecessors are found.

        NOTE(harlowja): predecessor cycles (if they exist) will not be iterated
        over more than once (this prevents infinite iteration).
        """
        visited = set([n])
        queue = collections.deque(self.predecessors(n))
        while queue:
            pred = queue.popleft()
            if pred not in visited:
                yield pred
                visited.add(pred)
                for pred_pred in self.predecessors(pred):
                    if pred_pred not in visited:
                        queue.append(pred_pred)

    def add_edge(self, u, v, attr_dict=None, **attr):
        """Add an edge between u and v."""
        if attr_dict is not None:
            return super(DiGraph, self).add_edge(u, v, **attr_dict)
        return super(DiGraph, self).add_edge(u, v, **attr)

    def add_node(self, n, attr_dict=None, **attr):
        """Add a single node n and update node attributes."""
        if attr_dict is not None:
            return super(DiGraph, self).add_node(n, **attr_dict)
        return super(DiGraph, self).add_node(n, **attr)

    def fresh_copy(self):
        """Return a fresh copy graph with the same data structure.

        A fresh copy has no nodes, edges or graph attributes. It is
        the same data structure as the current graph. This method is
        typically used to create an empty version of the graph.
        """
        return DiGraph()


class OrderedDiGraph(DiGraph):
    """A directed graph subclass with useful utility functions.

    This derivative retains node, edge, insertion and iteration
    ordering (so that the iteration order matches the insertion
    order).
    """
    node_dict_factory = collections.OrderedDict
    adjlist_outer_dict_factory = collections.OrderedDict
    adjlist_inner_dict_factory = collections.OrderedDict
    edge_attr_dict_factory = collections.OrderedDict

    def fresh_copy(self):
        """Return a fresh copy graph with the same data structure.

        A fresh copy has no nodes, edges or graph attributes. It is
        the same data structure as the current graph. This method is
        typically used to create an empty version of the graph.
        """
        return OrderedDiGraph()


class OrderedGraph(Graph):
    """A graph subclass with useful utility functions.

    This derivative retains node, edge, insertion and iteration
    ordering (so that the iteration order matches the insertion
    order).
    """
    node_dict_factory = collections.OrderedDict
    adjlist_outer_dict_factory = collections.OrderedDict
    adjlist_inner_dict_factory = collections.OrderedDict
    edge_attr_dict_factory = collections.OrderedDict

    def fresh_copy(self):
        """Return a fresh copy graph with the same data structure.

        A fresh copy has no nodes, edges or graph attributes. It is
        the same data structure as the current graph. This method is
        typically used to create an empty version of the graph.
        """
        return OrderedGraph()


def merge_graphs(graph, *graphs, **kwargs):
    """Merges a bunch of graphs into a new graph.

    If no additional graphs are provided the first graph is
    returned unmodified otherwise the merged graph is returned.
    """
    tmp_graph = graph
    allow_overlaps = kwargs.get('allow_overlaps', False)
    overlap_detector = kwargs.get('overlap_detector')
    if overlap_detector is not None and not callable(overlap_detector):
        raise ValueError("Overlap detection callback expected to be callable")
    elif overlap_detector is None:
        overlap_detector = (lambda to_graph, from_graph:
                            len(to_graph.subgraph(from_graph.nodes)))
    for g in graphs:
        # This should ensure that the nodes to be merged do not already exist
        # in the graph that is to be merged into. This could be problematic if
        # there are duplicates.
        if not allow_overlaps:
            # Attempt to induce a subgraph using the to be merged graphs nodes
            # and see if any graph results.
            overlaps = overlap_detector(graph, g)
            if overlaps:
                raise ValueError("Can not merge graph %s into %s since there "
                                 "are %s overlapping nodes (and we do not "
                                 "support merging nodes)" % (g, graph,
                                                             overlaps))
        graph = nx.algorithms.compose(graph, g)
    # Keep the first graphs name.
    if graphs:
        graph.name = tmp_graph.name
    return graph



# -*- coding: utf-8 -*-

#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
import threading

from oslo_utils import reflection
from oslo_utils import timeutils

from taskflow.engines.worker_based import protocol as pr
from taskflow import logging
from taskflow.utils import kombu_utils as ku

LOG = logging.getLogger(__name__)


# TODO(harlowja): this needs to be made better, once
# https://blueprints.launchpad.net/taskflow/+spec/wbe-worker-info is finally
# implemented we can go about using that instead.
class TopicWorker(object):
    """A (read-only) worker and its relevant information + useful methods."""

    _NO_IDENTITY = object()

    def __init__(self, topic, tasks, identity=_NO_IDENTITY):
        self.tasks = []
        for task in tasks:
            if not isinstance(task, str):
                task = reflection.get_class_name(task)
            self.tasks.append(task)
        self.topic = topic
        self.identity = identity
        self.last_seen = None

    def performs(self, task):
        if not isinstance(task, str):
            task = reflection.get_class_name(task)
        return task in self.tasks

    def __eq__(self, other):
        if not isinstance(other, TopicWorker):
            return NotImplemented
        if len(other.tasks) != len(self.tasks):
            return False
        if other.topic != self.topic:
            return False
        for task in other.tasks:
            if not self.performs(task):
                return False
        # If one of the identity equals _NO_IDENTITY, then allow it to match...
        if self._NO_IDENTITY in (self.identity, other.identity):
            return True
        else:
            return other.identity == self.identity

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        r = reflection.get_class_name(self, fully_qualified=False)
        if self.identity is not self._NO_IDENTITY:
            r += "(identity=%s, tasks=%s, topic=%s)" % (self.identity,
                                                        self.tasks, self.topic)
        else:
            r += "(identity=*, tasks=%s, topic=%s)" % (self.tasks, self.topic)
        return r


class ProxyWorkerFinder(object):
    """Requests and receives responses about workers topic+task details."""

    def __init__(self, uuid, proxy, topics,
                 beat_periodicity=pr.NOTIFY_PERIOD,
                 worker_expiry=pr.EXPIRES_AFTER):
        self._cond = threading.Condition()
        self._proxy = proxy
        self._topics = topics
        self._workers = {}
        self._uuid = uuid
        self._seen_workers = 0
        self._messages_processed = 0
        self._messages_published = 0
        self._worker_expiry = worker_expiry
        self._watch = timeutils.StopWatch(duration=beat_periodicity)

    @property
    def total_workers(self):
        """Number of workers currently known."""
        return len(self._workers)

    def wait_for_workers(self, workers=1, timeout=None):
        """Waits for geq workers to notify they are ready to do work.

        NOTE(harlowja): if a timeout is provided this function will wait
        until that timeout expires, if the amount of workers does not reach
        the desired amount of workers before the timeout expires then this will
        return how many workers are still needed, otherwise it will
        return zero.
        """
        if workers <= 0:
            raise ValueError("Worker amount must be greater than zero")
        watch = timeutils.StopWatch(duration=timeout)
        watch.start()
        with self._cond:
            while self.total_workers < workers:
                if watch.expired():
                    return max(0, workers - self.total_workers)
                self._cond.wait(watch.leftover(return_none=True))
            return 0

    @staticmethod
    def _match_worker(task, available_workers):
        """Select a worker (from geq 1 workers) that can best perform the task.

        NOTE(harlowja): this method will be activated when there exists
        one one greater than one potential workers that can perform a task,
        the arguments provided will be the potential workers located and the
        task that is being requested to perform and the result should be one
        of those workers using whatever best-fit algorithm is possible (or
        random at the least).
        """
        if len(available_workers) == 1:
            return available_workers[0]
        else:
            return random.choice(available_workers)

    @property
    def messages_processed(self):
        """How many notify response messages have been processed."""
        return self._messages_processed

    def _next_worker(self, topic, tasks, temporary=False):
        if not temporary:
            w = TopicWorker(topic, tasks, identity=self._seen_workers)
            self._seen_workers += 1
            return w
        else:
            return TopicWorker(topic, tasks)

    def maybe_publish(self):
        """Periodically called to publish notify message to each topic.

        These messages (especially the responses) are how this find learns
        about workers and what tasks they can perform (so that we can then
        match workers to tasks to run).
        """
        if self._messages_published == 0:
            self._proxy.publish(pr.Notify(),
                                self._topics, reply_to=self._uuid)
            self._messages_published += 1
            self._watch.restart()
        else:
            if self._watch.expired():
                self._proxy.publish(pr.Notify(),
                                    self._topics, reply_to=self._uuid)
                self._messages_published += 1
                self._watch.restart()

    def _add(self, topic, tasks):
        """Adds/updates a worker for the topic for the given tasks."""
        try:
            worker = self._workers[topic]
            # Check if we already have an equivalent worker, if so just
            # return it...
            if worker == self._next_worker(topic, tasks, temporary=True):
                return (worker, False)
            # This *fall through* is done so that if someone is using an
            # active worker object that already exists that we just create
            # a new one; so that the existing object doesn't get
            # affected (workers objects are supposed to be immutable).
        except KeyError:
            pass
        worker = self._next_worker(topic, tasks)
        self._workers[topic] = worker
        return (worker, True)

    def process_response(self, data, message):
        """Process notify message sent from remote side."""
        LOG.debug("Started processing notify response message '%s'",
                  ku.DelayedPretty(message))
        response = pr.Notify(**data)
        LOG.debug("Extracted notify response '%s'", response)
        with self._cond:
            worker, new_or_updated = self._add(response.topic,
                                               response.tasks)
            if new_or_updated:
                LOG.debug("Updated worker '%s' (%s total workers are"
                          " currently known)", worker, self.total_workers)
                self._cond.notify_all()
            worker.last_seen = timeutils.now()
            self._messages_processed += 1

    def clean(self):
        """Cleans out any dead/expired/not responding workers.

        Returns how many workers were removed.
        """
        if (not self._workers or
                (self._worker_expiry is None or self._worker_expiry <= 0)):
            return 0
        dead_workers = {}
        with self._cond:
            now = timeutils.now()
            for topic, worker in self._workers.items():
                if worker.last_seen is None:
                    continue
                secs_since_last_seen = max(0, now - worker.last_seen)
                if secs_since_last_seen >= self._worker_expiry:
                    dead_workers[topic] = (worker, secs_since_last_seen)
            for topic in dead_workers.keys():
                self._workers.pop(topic)
            if dead_workers:
                self._cond.notify_all()
        if dead_workers and LOG.isEnabledFor(logging.INFO):
            for worker, secs_since_last_seen in dead_workers.values():
                LOG.info("Removed worker '%s' as it has not responded to"
                         " notification requests in %0.3f seconds",
                         worker, secs_since_last_seen)
        return len(dead_workers)

    def reset(self):
        """Resets finders internal state."""
        with self._cond:
            self._workers.clear()
            self._messages_processed = 0
            self._messages_published = 0
            self._seen_workers = 0
            self._cond.notify_all()

    def get_worker_for_task(self, task):
        """Gets a worker that can perform a given task."""
        available_workers = []
        with self._cond:
            for worker in self._workers.values():
                if worker.performs(task):
                    available_workers.append(worker)
        if available_workers:
            return self._match_worker(task, available_workers)
        else:
            return None



# -*- coding: utf-8 -*-

#    Copyright (C) 2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import hashlib
import logging
import os
import random
import sys
import time

logging.basicConfig(level=logging.ERROR)

self_dir = os.path.abspath(os.path.dirname(__file__))
top_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                       os.pardir,
                                       os.pardir))
sys.path.insert(0, top_dir)
sys.path.insert(0, self_dir)

from oslo_utils import uuidutils

from taskflow import engines
from taskflow.patterns import graph_flow as gf
from taskflow.patterns import linear_flow as lf
from taskflow.persistence import models
from taskflow import task

import example_utils  # noqa

# INTRO: These examples show how a hierarchy of flows can be used to create a
# pseudo-volume in a reliable & resumable manner using taskflow + a miniature
# version of what cinder does while creating a volume (very miniature).


@contextlib.contextmanager
def slow_down(how_long=0.5):
    try:
        yield how_long
    finally:
        print("** Ctrl-c me please!!! **")
        time.sleep(how_long)


def find_flow_detail(backend, book_id, flow_id):
    # NOTE(harlowja): this is used to attempt to find a given logbook with
    # a given id and a given flow details inside that logbook, we need this
    # reference so that we can resume the correct flow (as a logbook tracks
    # flows and a flow detail tracks a individual flow).
    #
    # Without a reference to the logbook and the flow details in that logbook
    # we will not know exactly what we should resume and that would mean we
    # can't resume what we don't know.
    with contextlib.closing(backend.get_connection()) as conn:
        lb = conn.get_logbook(book_id)
        return lb.find(flow_id)


class PrintText(task.Task):
    def __init__(self, print_what, no_slow=False):
        content_hash = hashlib.md5(print_what.encode('utf-8')).hexdigest()[0:8]
        super(PrintText, self).__init__(name="Print: %s" % (content_hash))
        self._text = print_what
        self._no_slow = no_slow

    def execute(self):
        if self._no_slow:
            print("-" * (len(self._text)))
            print(self._text)
            print("-" * (len(self._text)))
        else:
            with slow_down():
                print("-" * (len(self._text)))
                print(self._text)
                print("-" * (len(self._text)))


class CreateSpecForVolumes(task.Task):
    def execute(self):
        volumes = []
        for i in range(0, random.randint(1, 10)):
            volumes.append({
                'type': 'disk',
                'location': "/dev/vda%s" % (i + 1),
            })
        return volumes


class PrepareVolumes(task.Task):
    def execute(self, volume_specs):
        for v in volume_specs:
            with slow_down():
                print("Dusting off your hard drive %s" % (v))
            with slow_down():
                print("Taking a well deserved break.")
            print("Your drive %s has been certified." % (v))


# Setup the set of things to do (mini-cinder).
flow = lf.Flow("root").add(
    PrintText("Starting volume create", no_slow=True),
    gf.Flow('maker').add(
        CreateSpecForVolumes("volume_specs", provides='volume_specs'),
        PrintText("I need a nap, it took me a while to build those specs."),
        PrepareVolumes(),
    ),
    PrintText("Finished volume create", no_slow=True))

# Setup the persistence & resumption layer.
with example_utils.get_backend() as backend:
    try:
        book_id, flow_id = sys.argv[2].split("+", 1)
    except (IndexError, ValueError):
        book_id = None
        flow_id = None

    if not all([book_id, flow_id]):
        # If no 'tracking id' (think a fedex or ups tracking id) is provided
        # then we create one by creating a logbook (where flow details are
        # stored) and creating a flow detail (where flow and task state is
        # stored). The combination of these 2 objects unique ids (uuids) allows
        # the users of taskflow to reassociate the workflows that were
        # potentially running (and which may have partially completed) back
        # with taskflow so that those workflows can be resumed (or reverted)
        # after a process/thread/engine has failed in someway.
        book = models.LogBook('resume-volume-create')
        flow_detail = models.FlowDetail("root", uuid=uuidutils.generate_uuid())
        book.add(flow_detail)
        with contextlib.closing(backend.get_connection()) as conn:
            conn.save_logbook(book)
        print("!! Your tracking id is: '%s+%s'" % (book.uuid,
                                                   flow_detail.uuid))
        print("!! Please submit this on later runs for tracking purposes")
    else:
        flow_detail = find_flow_detail(backend, book_id, flow_id)

    # Load and run.
    engine = engines.load(flow,
                          flow_detail=flow_detail,
                          backend=backend, engine='serial')
    engine.run()

# How to use.
#
# 1. $ python me.py "sqlite:////tmp/cinder.db"
# 2. ctrl-c before this finishes
# 3. Find the tracking id (search for 'Your tracking id is')
# 4. $ python me.py "sqlite:////tmp/cinder.db" "$tracking_id"
# 5. Profit!



# -*- coding: utf-8 -*-

#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import string

from taskflow import test
from taskflow.utils import iter_utils


def forever_it():
    i = 0
    while True:
        yield i
        i += 1


class IterUtilsTest(test.TestCase):
    def test_fill_empty(self):
        self.assertEqual([], list(iter_utils.fill([1, 2, 3], 0)))

    def test_bad_unique_seen(self):
        iters = [
            ['a', 'b'],
            2,
            None,
            object(),
        ]
        self.assertRaises(ValueError,
                          iter_utils.unique_seen, iters)

    def test_generate_delays(self):
        it = iter_utils.generate_delays(1, 60)
        self.assertEqual(1, next(it))
        self.assertEqual(2, next(it))
        self.assertEqual(4, next(it))
        self.assertEqual(8, next(it))
        self.assertEqual(16, next(it))
        self.assertEqual(32, next(it))
        self.assertEqual(60, next(it))
        self.assertEqual(60, next(it))

    def test_generate_delays_custom_multiplier(self):
        it = iter_utils.generate_delays(1, 60, multiplier=4)
        self.assertEqual(1, next(it))
        self.assertEqual(4, next(it))
        self.assertEqual(16, next(it))
        self.assertEqual(60, next(it))
        self.assertEqual(60, next(it))

    def test_generate_delays_bad(self):
        self.assertRaises(ValueError, iter_utils.generate_delays, -1, -1)
        self.assertRaises(ValueError, iter_utils.generate_delays, -1, 2)
        self.assertRaises(ValueError, iter_utils.generate_delays, 2, -1)
        self.assertRaises(ValueError, iter_utils.generate_delays, 1, 1,
                          multiplier=0.5)

    def test_unique_seen(self):
        iters = [
            ['a', 'b'],
            ['a', 'c', 'd'],
            ['a', 'e', 'f'],
            ['f', 'm', 'n'],
        ]
        self.assertEqual(['a', 'b', 'c', 'd', 'e', 'f', 'm', 'n'],
                         list(iter_utils.unique_seen(iters)))

    def test_unique_seen_empty(self):
        iters = []
        self.assertEqual([], list(iter_utils.unique_seen(iters)))

    def test_unique_seen_selector(self):
        iters = [
            [(1, 'a'), (1, 'a')],
            [(2, 'b')],
            [(3, 'c')],
            [(1, 'a'), (3, 'c')],
        ]
        it = iter_utils.unique_seen(iters,
                                    seen_selector=lambda value: value[0])
        self.assertEqual([(1, 'a'), (2, 'b'), (3, 'c')], list(it))

    def test_bad_fill(self):
        self.assertRaises(ValueError, iter_utils.fill, 2, 2)

    def test_fill_many_empty(self):
        result = list(iter_utils.fill(range(0, 50), 500))
        self.assertEqual(450, sum(1 for x in result if x is None))
        self.assertEqual(50, sum(1 for x in result if x is not None))

    def test_fill_custom_filler(self):
        self.assertEqual("abcd",
                         "".join(iter_utils.fill("abc", 4, filler='d')))

    def test_fill_less_needed(self):
        self.assertEqual("ab", "".join(iter_utils.fill("abc", 2)))

    def test_fill(self):
        self.assertEqual([None, None], list(iter_utils.fill([], 2)))
        self.assertEqual((None, None), tuple(iter_utils.fill([], 2)))

    def test_bad_find_first_match(self):
        self.assertRaises(ValueError,
                          iter_utils.find_first_match, 2, lambda v: False)

    def test_find_first_match(self):
        it = forever_it()
        self.assertEqual(100, iter_utils.find_first_match(it,
                                                          lambda v: v == 100))

    def test_find_first_match_not_found(self):
        it = iter(string.ascii_lowercase)
        self.assertIsNone(iter_utils.find_first_match(it,
                                                      lambda v: v == ''))

    def test_bad_count(self):
        self.assertRaises(ValueError, iter_utils.count, 2)

    def test_count(self):
        self.assertEqual(0, iter_utils.count([]))
        self.assertEqual(1, iter_utils.count(['a']))
        self.assertEqual(10, iter_utils.count(range(0, 10)))
        self.assertEqual(1000, iter_utils.count(range(0, 1000)))
        self.assertEqual(0, iter_utils.count(range(0)))
        self.assertEqual(0, iter_utils.count(range(-1)))

    def test_bad_while_is_not(self):
        self.assertRaises(ValueError, iter_utils.while_is_not, 2, 'a')

    def test_while_is_not(self):
        class Dummy(object):
            def __init__(self, char):
                self.char = char
        dummy_list = [Dummy(a)
                      for a in string.ascii_lowercase]

        it = iter(dummy_list)
        self.assertEqual([dummy_list[0]],
                         list(iter_utils.while_is_not(it, dummy_list[0])))
        it = iter(dummy_list)
        self.assertEqual(dummy_list[0:2],
                         list(iter_utils.while_is_not(it, dummy_list[1])))
        self.assertEqual(dummy_list[2:],
                         list(iter_utils.while_is_not(it, Dummy('zzz'))))

        it = iter(dummy_list)
        self.assertEqual(dummy_list,
                         list(iter_utils.while_is_not(it, Dummy(''))))



# -*- coding: utf-8 -*-

#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading


class Timeout(object):
    """An object which represents a timeout.

    This object has the ability to be interrupted before the actual timeout
    is reached.
    """
    def __init__(self, value, event_factory=threading.Event):
        if value < 0:
            raise ValueError("Timeout value must be greater or"
                             " equal to zero and not '%s'" % (value))
        self._value = value
        self._event = event_factory()

    @property
    def value(self):
        """Immutable value of the internally used timeout."""
        return self._value

    def interrupt(self):
        """Forcefully set the timeout (releases any waiters)."""
        self._event.set()

    def is_stopped(self):
        """Returns if the timeout has been interrupted."""
        return self._event.is_set()

    def wait(self):
        """Block current thread (up to timeout) and wait until interrupted."""
        self._event.wait(self._value)

    def reset(self):
        """Reset so that interruption (and waiting) can happen again."""
        self._event.clear()


def convert_to_timeout(value=None, default_value=None,
                       event_factory=threading.Event):
    """Converts a given value to a timeout instance (and returns it).

    Does nothing if the value provided is already a timeout instance.
    """
    if value is None:
        value = default_value
    if isinstance(value, (int, float, str)):
        return Timeout(float(value), event_factory=event_factory)
    elif isinstance(value, Timeout):
        return value
    else:
        raise ValueError("Invalid timeout literal '%s'" % (value))



# -*- coding: utf-8 -*-

#    Copyright (C) 2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import futurist

from taskflow import task as ta
from taskflow.types import failure
from taskflow.types import notifier

# Execution and reversion outcomes.
EXECUTED = 'executed'
REVERTED = 'reverted'


def _execute_retry(retry, arguments):
    try:
        result = retry.execute(**arguments)
    except Exception:
        result = failure.Failure()
    return (EXECUTED, result)


def _revert_retry(retry, arguments):
    try:
        result = retry.revert(**arguments)
    except Exception:
        result = failure.Failure()
    return (REVERTED, result)


def _execute_task(task, arguments, progress_callback=None):
    with notifier.register_deregister(task.notifier,
                                      ta.EVENT_UPDATE_PROGRESS,
                                      callback=progress_callback):
        try:
            task.pre_execute()
            result = task.execute(**arguments)
        except Exception:
            # NOTE(imelnikov): wrap current exception with Failure
            # object and return it.
            result = failure.Failure()
        finally:
            task.post_execute()
    return (EXECUTED, result)


def _revert_task(task, arguments, result, failures, progress_callback=None):
    arguments = arguments.copy()
    arguments[ta.REVERT_RESULT] = result
    arguments[ta.REVERT_FLOW_FAILURES] = failures
    with notifier.register_deregister(task.notifier,
                                      ta.EVENT_UPDATE_PROGRESS,
                                      callback=progress_callback):
        try:
            task.pre_revert()
            result = task.revert(**arguments)
        except Exception:
            # NOTE(imelnikov): wrap current exception with Failure
            # object and return it.
            result = failure.Failure()
        finally:
            task.post_revert()
    return (REVERTED, result)


class SerialRetryExecutor(object):
    """Executes and reverts retries."""

    def __init__(self):
        self._executor = futurist.SynchronousExecutor()

    def start(self):
        """Prepare to execute retries."""
        self._executor.restart()

    def stop(self):
        """Finalize retry executor."""
        self._executor.shutdown()

    def execute_retry(self, retry, arguments):
        """Schedules retry execution."""
        fut = self._executor.submit(_execute_retry, retry, arguments)
        fut.atom = retry
        return fut

    def revert_retry(self, retry, arguments):
        """Schedules retry reversion."""
        fut = self._executor.submit(_revert_retry, retry, arguments)
        fut.atom = retry
        return fut


class TaskExecutor(object, metaclass=abc.ABCMeta):
    """Executes and reverts tasks.

    This class takes task and its arguments and executes or reverts it.
    It encapsulates knowledge on how task should be executed or reverted:
    right now, on separate thread, on another machine, etc.
    """

    @abc.abstractmethod
    def execute_task(self, task, task_uuid, arguments,
                     progress_callback=None):
        """Schedules task execution."""

    @abc.abstractmethod
    def revert_task(self, task, task_uuid, arguments, result, failures,
                    progress_callback=None):
        """Schedules task reversion."""

    def start(self):
        """Prepare to execute tasks."""

    def stop(self):
        """Finalize task executor."""


class SerialTaskExecutor(TaskExecutor):
    """Executes tasks one after another."""

    def __init__(self):
        self._executor = futurist.SynchronousExecutor()

    def start(self):
        self._executor.restart()

    def stop(self):
        self._executor.shutdown()

    def execute_task(self, task, task_uuid, arguments, progress_callback=None):
        fut = self._executor.submit(_execute_task,
                                    task, arguments,
                                    progress_callback=progress_callback)
        fut.atom = task
        return fut

    def revert_task(self, task, task_uuid, arguments, result, failures,
                    progress_callback=None):
        fut = self._executor.submit(_revert_task,
                                    task, arguments, result, failures,
                                    progress_callback=progress_callback)
        fut.atom = task
        return fut


class ParallelTaskExecutor(TaskExecutor):
    """Executes tasks in parallel.

    Submits tasks to an executor which should provide an interface similar
    to concurrent.Futures.Executor.
    """

    constructor_options = [
        ('max_workers', lambda v: v if v is None else int(v)),
    ]
    """
    Optional constructor keyword arguments this executor supports. These will
    typically be passed via engine options (by a engine user) and converted
    into the correct type before being sent into this
    classes ``__init__`` method.
    """

    def __init__(self, executor=None, max_workers=None):
        self._executor = executor
        self._max_workers = max_workers
        self._own_executor = executor is None

    @abc.abstractmethod
    def _create_executor(self, max_workers=None):
        """Called when an executor has not been provided to make one."""

    def _submit_task(self, func, task, *args, **kwargs):
        fut = self._executor.submit(func, task, *args, **kwargs)
        fut.atom = task
        return fut

    def execute_task(self, task, task_uuid, arguments, progress_callback=None):
        return self._submit_task(_execute_task, task, arguments,
                                 progress_callback=progress_callback)

    def revert_task(self, task, task_uuid, arguments, result, failures,
                    progress_callback=None):
        return self._submit_task(_revert_task, task, arguments, result,
                                 failures, progress_callback=progress_callback)

    def start(self):
        if self._own_executor:
            self._executor = self._create_executor(
                max_workers=self._max_workers)

    def stop(self):
        if self._own_executor:
            self._executor.shutdown(wait=True)
            self._executor = None


class ParallelThreadTaskExecutor(ParallelTaskExecutor):
    """Executes tasks in parallel using a thread pool executor."""

    def _create_executor(self, max_workers=None):
        return futurist.ThreadPoolExecutor(max_workers=max_workers)


class ParallelGreenThreadTaskExecutor(ParallelThreadTaskExecutor):
    """Executes tasks in parallel using a greenthread pool executor."""

    DEFAULT_WORKERS = 1000
    """
    Default number of workers when ``None`` is passed; being that
    greenthreads don't map to native threads or processors very well this
    is more of a guess/somewhat arbitrary, but it does match what the eventlet
    greenpool default size is (so at least it's consistent with what eventlet
    does).
    """

    def _create_executor(self, max_workers=None):
        if max_workers is None:
            max_workers = self.DEFAULT_WORKERS
        return futurist.GreenThreadPoolExecutor(max_workers=max_workers)



#!/usr/bin/env python3

#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import optparse
import os
import sys

top_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                       os.pardir))
sys.path.insert(0, top_dir)

from automaton.converters import pydot
from automaton import machines

from taskflow.engines.action_engine import builder
from taskflow.engines.worker_based import protocol
from taskflow import states


# This is just needed to get at the machine object (we will not
# actually be running it...).
class DummyRuntime(object):
    def __init__(self):
        self.analyzer = mock.MagicMock()
        self.completer = mock.MagicMock()
        self.scheduler = mock.MagicMock()
        self.storage = mock.MagicMock()
        self.selector = mock.MagicMock()


def make_machine(start_state, transitions, event_name_cb):
    machine = machines.FiniteMachine()
    machine.add_state(start_state)
    machine.default_start_state = start_state
    for (start_state, end_state) in transitions:
        if start_state not in machine:
            machine.add_state(start_state)
        if end_state not in machine:
            machine.add_state(end_state)
        event = event_name_cb(start_state, end_state)
        machine.add_transition(start_state, end_state, event)
    return machine


def main():
    parser = optparse.OptionParser()
    parser.add_option("-f", "--file", dest="filename",
                      help="write svg to FILE", metavar="FILE")
    parser.add_option("-t", "--tasks", dest="tasks",
                      action='store_true',
                      help="use task state transitions",
                      default=False)
    parser.add_option("-r", "--retries", dest="retries",
                      action='store_true',
                      help="use retry state transitions",
                      default=False)
    parser.add_option("-e", "--engines", dest="engines",
                      action='store_true',
                      help="use engine state transitions",
                      default=False)
    parser.add_option("-w", "--wbe-requests", dest="wbe_requests",
                      action='store_true',
                      help="use wbe request transitions",
                      default=False)
    parser.add_option("-j", "--jobs", dest="jobs",
                      action='store_true',
                      help="use job transitions",
                      default=False)
    parser.add_option("--flow", dest="flow",
                      action='store_true',
                      help="use flow transitions",
                      default=False)
    parser.add_option("-T", "--format", dest="format",
                      help="output in given format",
                      default='svg')

    (options, args) = parser.parse_args()
    if options.filename is None:
        options.filename = 'states.%s' % options.format

    types = [
        options.engines,
        options.retries,
        options.tasks,
        options.wbe_requests,
        options.jobs,
        options.flow,
    ]
    provided = sum([int(i) for i in types])
    if provided > 1:
        parser.error("Only one of task/retry/engines/wbe requests/jobs/flow"
                     " may be specified.")
    if provided == 0:
        parser.error("One of task/retry/engines/wbe requests/jobs/flow"
                     " must be specified.")

    event_name_cb = lambda start_state, end_state: "on_%s" % end_state.lower()
    internal_states = list()
    ordering = 'in'
    if options.tasks:
        source_type = "Tasks"
        source = make_machine(states.PENDING,
                              list(states._ALLOWED_TASK_TRANSITIONS),
                              event_name_cb)
    elif options.retries:
        source_type = "Retries"
        source = make_machine(states.PENDING,
                              list(states._ALLOWED_RETRY_TRANSITIONS),
                              event_name_cb)
    elif options.flow:
        source_type = "Flow"
        source = make_machine(states.PENDING,
                              list(states._ALLOWED_FLOW_TRANSITIONS),
                              event_name_cb)
    elif options.engines:
        source_type = "Engines"
        b = builder.MachineBuilder(DummyRuntime(), mock.MagicMock())
        source, memory = b.build({})
        internal_states.extend(builder.META_STATES)
        ordering = 'out'
    elif options.wbe_requests:
        source_type = "WBE requests"
        source = make_machine(protocol.WAITING,
                              list(protocol._ALLOWED_TRANSITIONS),
                              event_name_cb)
    elif options.jobs:
        source_type = "Jobs"
        source = make_machine(states.UNCLAIMED,
                              list(states._ALLOWED_JOB_TRANSITIONS),
                              event_name_cb)

    graph_attrs = {
        'ordering': ordering,
    }
    graph_name = "%s states" % source_type

    def node_attrs_cb(state):
        node_color = None
        if state in internal_states:
            node_color = 'blue'
        if state in (states.FAILURE, states.REVERT_FAILURE):
            node_color = 'red'
        if state == states.REVERTED:
            node_color = 'darkorange'
        if state in (states.SUCCESS, states.COMPLETE):
            node_color = 'green'
        node_attrs = {}
        if node_color:
            node_attrs['fontcolor'] = node_color
        return node_attrs

    def edge_attrs_cb(start_state, on_event, end_state):
        edge_attrs = {}
        if options.engines:
            edge_attrs['label'] = on_event.replace("_", " ").strip()
            if 'reverted' in on_event:
                edge_attrs['fontcolor'] = 'darkorange'
            if 'fail' in on_event:
                edge_attrs['fontcolor'] = 'red'
            if 'success' in on_event:
                edge_attrs['fontcolor'] = 'green'
        return edge_attrs

    g = pydot.convert(source, graph_name, graph_attrs=graph_attrs,
                      node_attrs_cb=node_attrs_cb, edge_attrs_cb=edge_attrs_cb)
    print("*" * len(graph_name))
    print(graph_name)
    print("*" * len(graph_name))
    print(source.pformat())
    print(g.to_string().strip())

    g.write(options.filename, format=options.format)
    print("Created %s at '%s'" % (options.format, options.filename))

    # To make the svg more pretty use the following:
    # $ xsltproc ../diagram-tools/notugly.xsl ./states.svg > pretty-states.svg
    # Get diagram-tools from https://github.com/vidarh/diagram-tools.git


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#    Copyright 2015 Hewlett-Packard Development Company, L.P.
#    Copyright (C) 2013 Rackspace Hosting Inc. All Rights Reserved.
#    Copyright (C) 2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import copy
import functools

from oslo_utils import reflection

from taskflow import atom
from taskflow import logging
from taskflow.types import notifier
from taskflow.utils import misc

LOG = logging.getLogger(__name__)

# Constants passed into revert kwargs.
#
# Contain the execute() result (if any).
REVERT_RESULT = 'result'
#
# The cause of the flow failure/s
REVERT_FLOW_FAILURES = 'flow_failures'

# Common events
EVENT_UPDATE_PROGRESS = 'update_progress'


class Task(atom.Atom, metaclass=abc.ABCMeta):
    """An abstraction that defines a potential piece of work.

    This potential piece of work is expected to be able to contain
    functionality that defines what can be executed to accomplish that work
    as well as a way of defining what can be executed to reverted/undo that
    same piece of work.
    """

    # Known internal events this task can have callbacks bound to (others that
    # are not in this set/tuple will not be able to be bound); this should be
    # updated and/or extended in subclasses as needed to enable or disable new
    # or existing internal events...
    TASK_EVENTS = (EVENT_UPDATE_PROGRESS,)

    def __init__(self, name=None, provides=None, requires=None,
                 auto_extract=True, rebind=None, inject=None,
                 ignore_list=None, revert_rebind=None, revert_requires=None):
        if name is None:
            name = reflection.get_class_name(self)
        super(Task, self).__init__(name, provides=provides, requires=requires,
                                   auto_extract=auto_extract, rebind=rebind,
                                   inject=inject, revert_rebind=revert_rebind,
                                   revert_requires=revert_requires)
        self._notifier = notifier.RestrictedNotifier(self.TASK_EVENTS)

    @property
    def notifier(self):
        """Internal notification dispatcher/registry.

        A notification object that will dispatch events that occur related
        to *internal* notifications that the task internally emits to
        listeners (for example for progress status updates, telling others
        that a task has reached 50% completion...).
        """
        return self._notifier

    def copy(self, retain_listeners=True):
        """Clone/copy this task.

        :param retain_listeners: retain the attached notification listeners
                                 when cloning, when false the listeners will
                                 be emptied, when true the listeners will be
                                 copied and retained

        :return: the copied task
        """
        c = copy.copy(self)
        c._notifier = self._notifier.copy()
        if not retain_listeners:
            c._notifier.reset()
        return c

    def update_progress(self, progress):
        """Update task progress and notify all registered listeners.

        :param progress: task progress float value between 0.0 and 1.0
        """
        def on_clamped():
            LOG.warning("Progress value must be greater or equal to 0.0 or"
                        " less than or equal to 1.0 instead of being '%s'",
                        progress)
        cleaned_progress = misc.clamp(progress, 0.0, 1.0,
                                      on_clamped=on_clamped)
        self._notifier.notify(EVENT_UPDATE_PROGRESS,
                              {'progress': cleaned_progress})


class FunctorTask(Task):
    """Adaptor to make a task from a callable.

    Take any callable pair and make a task from it.

    NOTE(harlowja): If a name is not provided the function/method name of
    the ``execute`` callable will be used as the name instead (the name of
    the ``revert`` callable is not used).
    """

    def __init__(self, execute, name=None, provides=None,
                 requires=None, auto_extract=True, rebind=None, revert=None,
                 version=None, inject=None):
        if not callable(execute):
            raise ValueError("Function to use for executing must be"
                             " callable")
        if revert is not None:
            if not callable(revert):
                raise ValueError("Function to use for reverting must"
                                 " be callable")
        if name is None:
            name = reflection.get_callable_name(execute)
        super(FunctorTask, self).__init__(name, provides=provides,
                                          inject=inject)
        self._execute = execute
        self._revert = revert
        if version is not None:
            self.version = version
        mapping = self._build_arg_mapping(execute, requires, rebind,
                                          auto_extract)
        self.rebind, exec_requires, self.optional = mapping

        if revert:
            revert_mapping = self._build_arg_mapping(revert, requires, rebind,
                                                     auto_extract)
        else:
            revert_mapping = (self.rebind, exec_requires, self.optional)
        (self.revert_rebind, revert_requires,
         self.revert_optional) = revert_mapping
        self.requires = exec_requires.union(revert_requires)

    def execute(self, *args, **kwargs):
        return self._execute(*args, **kwargs)

    def revert(self, *args, **kwargs):
        if self._revert:
            return self._revert(*args, **kwargs)
        else:
            return None


class ReduceFunctorTask(Task):
    """General purpose Task to reduce a list by applying a function.

    This Task mimics the behavior of Python's built-in ``reduce`` function. The
    Task takes a functor (lambda or otherwise) and a list. The list is
    specified using the ``requires`` argument of the Task. When executed, this
    task calls ``reduce`` with the functor and list as arguments. The resulting
    value from the call to ``reduce`` is then returned after execution.
    """
    def __init__(self, functor, requires, name=None, provides=None,
                 auto_extract=True, rebind=None, inject=None):

        if not callable(functor):
            raise ValueError("Function to use for reduce must be callable")

        f_args = reflection.get_callable_args(functor)
        if len(f_args) != 2:
            raise ValueError("%s arguments were provided. Reduce functor "
                             "must take exactly 2 arguments." % len(f_args))

        if not misc.is_iterable(requires):
            raise TypeError("%s type was provided for requires. Requires "
                            "must be an iterable." % type(requires))

        if len(requires) < 2:
            raise ValueError("%s elements were provided. Requires must have "
                             "at least 2 elements." % len(requires))

        if name is None:
            name = reflection.get_callable_name(functor)
        super(ReduceFunctorTask, self).__init__(name=name,
                                                provides=provides,
                                                inject=inject,
                                                requires=requires,
                                                rebind=rebind,
                                                auto_extract=auto_extract)

        self._functor = functor

    def execute(self, *args, **kwargs):
        l = [kwargs[r] for r in self.requires]
        return functools.reduce(self._functor, l)


class MapFunctorTask(Task):
    """General purpose Task to map a function to a list.

    This Task mimics the behavior of Python's built-in ``map`` function. The
    Task takes a functor (lambda or otherwise) and a list. The list is
    specified using the ``requires`` argument of the Task. When executed, this
    task calls ``map`` with the functor and list as arguments. The resulting
    list from the call to ``map`` is then returned after execution.

    Each value of the returned list can be bound to individual names using
    the ``provides`` argument, following taskflow standard behavior. Order is
    preserved in the returned list.
    """

    def __init__(self, functor, requires, name=None, provides=None,
                 auto_extract=True, rebind=None, inject=None):

        if not callable(functor):
            raise ValueError("Function to use for map must be callable")

        f_args = reflection.get_callable_args(functor)
        if len(f_args) != 1:
            raise ValueError("%s arguments were provided. Map functor must "
                             "take exactly 1 argument." % len(f_args))

        if not misc.is_iterable(requires):
            raise TypeError("%s type was provided for requires. Requires "
                            "must be an iterable." % type(requires))

        if name is None:
            name = reflection.get_callable_name(functor)
        super(MapFunctorTask, self).__init__(name=name, provides=provides,
                                             inject=inject, requires=requires,
                                             rebind=rebind,
                                             auto_extract=auto_extract)

        self._functor = functor

    def execute(self, *args, **kwargs):
        l = [kwargs[r] for r in self.requires]
        return list(map(self._functor, l))



# -*- coding: utf-8 -*-

#    Copyright (C) 2014 Ivan Melnikov <iv at altlinux dot org>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os
import sys

logging.basicConfig(level=logging.ERROR)

top_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                       os.pardir,
                                       os.pardir))
sys.path.insert(0, top_dir)

import taskflow.engines
from taskflow.patterns import linear_flow as lf
from taskflow import task

# INTRO: pseudo-scoping by adding prefixes

# Sometimes you need scoping -- e.g. for adding several
# similar subflows to one flow to do same stuff for different
# data. But current version of TaskFlow does not allow that
# directly, so you have to resort to some kind of trickery.
# One (and more or less recommended, if not the only) way of
# solving the problem is to transform every task name, it's
# provides and requires values -- e.g. by adding prefix to them.
# This example shows how this could be done.


# The example task is simple: for each specified person, fetch
# his or her phone number from phone book and call.


PHONE_BOOK = {
    'jim': '444',
    'joe': '555',
    'iv_m': '666',
    'josh': '777'
}


class FetchNumberTask(task.Task):
    """Task that fetches number from phone book."""

    default_provides = 'number'

    def execute(self, person):
        print('Fetching number for %s.' % person)
        return PHONE_BOOK[person.lower()]


class CallTask(task.Task):
    """Task that calls person by number."""

    def execute(self, person, number):
        print('Calling %s %s.' % (person, number))

# This is how it works for one person:

simple_flow = lf.Flow('simple one').add(
    FetchNumberTask(),
    CallTask())
print('Running simple flow:')
taskflow.engines.run(simple_flow, store={'person': 'Josh'})


# To call several people you'll need a factory function that will
# make a flow with given prefix for you. We need to add prefix
# to task names, their provides and requires values. For requires,
# we use `rebind` argument of task constructor.
def subflow_factory(prefix):
    def pr(what):
        return '%s-%s' % (prefix, what)

    return lf.Flow(pr('flow')).add(
        FetchNumberTask(pr('fetch'),
                        provides=pr('number'),
                        rebind=[pr('person')]),
        CallTask(pr('call'),
                 rebind=[pr('person'), pr('number')])
    )


def call_them_all():
    # Let's call them all. We need a flow:
    flow = lf.Flow('call-them-prefixed')

    # We'll also need to inject person names with prefixed argument
    # name to storage to satisfy task requirements.
    persons = {}

    for person in ('Jim', 'Joe', 'Josh'):
        prefix = person.lower()
        persons['%s-person' % prefix] = person
        flow.add(subflow_factory(prefix))
    taskflow.engines.run(flow, store=persons)

print('\nCalling many people using prefixed factory:')
call_them_all()



# -*- coding: utf-8 -*-

#    Copyright (C) 2016 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import string

from taskflow.utils import misc
from taskflow import version


BANNER_HEADER = string.Template("""
___    __
 |    |_
 |ask |low v$version
""".strip())
BANNER_HEADER = BANNER_HEADER.substitute(version=version.version_string())


def make_banner(what, chapters):
    """Makes a taskflow banner string.

    For example::

      >>> from taskflow.utils import banner
      >>> chapters = {
          'Connection details': {
              'Topic': 'hello',
          },
          'Powered by': {
              'Executor': 'parallel',
          },
      }
      >>> print(banner.make_banner('Worker', chapters))

    This will output::

      ___    __
       |    |_
       |ask |low v1.26.1
      *Worker*
      Connection details:
        Topic => hello
      Powered by:
        Executor => parallel
    """
    buf = misc.StringIO()
    buf.write_nl(BANNER_HEADER)
    if chapters:
        buf.write_nl("*%s*" % what)
        chapter_names = sorted(chapters.keys())
    else:
        buf.write("*%s*" % what)
        chapter_names = []
    for i, chapter_name in enumerate(chapter_names):
        chapter_contents = chapters[chapter_name]
        if chapter_contents:
            buf.write_nl("%s:" % (chapter_name))
        else:
            buf.write("%s:" % (chapter_name))
        if isinstance(chapter_contents, dict):
            section_names = sorted(chapter_contents.keys())
            for j, section_name in enumerate(section_names):
                if j + 1 < len(section_names):
                    buf.write_nl("  %s => %s"
                                 % (section_name,
                                    chapter_contents[section_name]))
                else:
                    buf.write("  %s => %s" % (section_name,
                                              chapter_contents[section_name]))
        elif isinstance(chapter_contents, (list, tuple, set)):
            if isinstance(chapter_contents, set):
                sections = sorted(chapter_contents)
            else:
                sections = chapter_contents
            for j, section in enumerate(sections):
                if j + 1 < len(sections):
                    buf.write_nl("  %s. %s" % (j + 1, section))
                else:
                    buf.write("  %s. %s" % (j + 1, section))
        else:
            raise TypeError("Unsupported chapter contents"
                            " type: one of dict, list, tuple, set expected"
                            " and not %s" % type(chapter_contents).__name__)
        if i + 1 < len(chapter_names):
            buf.write_nl("")
    # NOTE(harlowja): this is needed since the template in this file
    # will always have newlines that end with '\n' (even on different
    # platforms due to the way this source file is encoded) so we have
    # to do this little dance to make it platform neutral...
    if os.linesep != "\n":
        return misc.fix_newlines(buf.getvalue())
    return buf.getvalue()



# -*- coding: utf-8 -*-

#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import os
import threading

import fasteners

from taskflow import engines
from taskflow import exceptions as excp
from taskflow.types import entity
from taskflow.types import notifier
from taskflow.utils import misc


class Conductor(object, metaclass=abc.ABCMeta):
    """Base for all conductor implementations.

    Conductors act as entities which extract jobs from a jobboard, assign
    there work to some engine (using some desired configuration) and then wait
    for that work to complete. If the work fails then they abandon the claimed
    work (or if the process they are running in crashes or dies this
    abandonment happens automatically) and then another conductor at a later
    period of time will finish up the prior failed conductors work.
    """

    #: Entity kind used when creating new entity objects
    ENTITY_KIND = 'conductor'

    def __init__(self, name, jobboard,
                 persistence=None, engine=None, engine_options=None):
        self._name = name
        self._jobboard = jobboard
        self._engine = engine
        self._engine_options = misc.safe_copy_dict(engine_options)
        self._persistence = persistence
        self._lock = threading.RLock()
        self._notifier = notifier.Notifier()

    @misc.cachedproperty
    def conductor(self):
        """Entity object that represents this conductor."""
        hostname = misc.get_hostname()
        pid = os.getpid()
        name = '@'.join([self._name, hostname + ":" + str(pid)])
        metadata = {
            'hostname': hostname,
            'pid': pid,
        }
        return entity.Entity(self.ENTITY_KIND, name, metadata)

    @property
    def notifier(self):
        """The conductor actions (or other state changes) notifier.

        NOTE(harlowja): different conductor implementations may emit
        different events + event details at different times, so refer to your
        conductor documentation to know exactly what can and what can not be
        subscribed to.
        """
        return self._notifier

    def _flow_detail_from_job(self, job):
        """Extracts a flow detail from a job (via some manner).

        The current mechanism to accomplish this is the following choices:

        * If the job details provide a 'flow_uuid' key attempt to load this
          key from the jobs book and use that as the flow_detail to run.
        * If the job details does not have have a 'flow_uuid' key then attempt
          to examine the size of the book and if it's only one element in the
          book (aka one flow_detail) then just use that.
        * Otherwise if there is no 'flow_uuid' defined or there are > 1
          flow_details in the book raise an error that corresponds to being
          unable to locate the correct flow_detail to run.
        """
        book = job.book
        if book is None:
            raise excp.NotFound("No book found in job")
        if job.details and 'flow_uuid' in job.details:
            flow_uuid = job.details["flow_uuid"]
            flow_detail = book.find(flow_uuid)
            if flow_detail is None:
                raise excp.NotFound("No matching flow detail found in"
                                    " jobs book for flow detail"
                                    " with uuid %s" % flow_uuid)
        else:
            choices = len(book)
            if choices == 1:
                flow_detail = list(book)[0]
            elif choices == 0:
                raise excp.NotFound("No flow detail(s) found in jobs book")
            else:
                raise excp.MultipleChoices("No matching flow detail found (%s"
                                           " choices) in jobs book" % choices)
        return flow_detail

    def _engine_from_job(self, job):
        """Extracts an engine from a job (via some manner)."""
        flow_detail = self._flow_detail_from_job(job)
        store = {}

        if flow_detail.meta and 'store' in flow_detail.meta:
            store.update(flow_detail.meta['store'])

        if job.details and 'store' in job.details:
            store.update(job.details["store"])

        engine = engines.load_from_detail(flow_detail, store=store,
                                          engine=self._engine,
                                          backend=self._persistence,
                                          **self._engine_options)
        return engine

    def _listeners_from_job(self, job, engine):
        """Returns a list of listeners to be attached to an engine.

        This method should be overridden in order to attach listeners to
        engines. It will be called once for each job, and the list returned
        listeners will be added to the engine for this job.

        :param job: A job instance that is about to be run in an engine.
        :param engine: The engine that listeners will be attached to.
        :returns: a list of (unregistered) listener instances.
        """
        # TODO(dkrause): Create a standard way to pass listeners or
        #                listener factories over the jobboard
        return []

    @fasteners.locked
    def connect(self):
        """Ensures the jobboard is connected (noop if it is already)."""
        if not self._jobboard.connected:
            self._jobboard.connect()

    @fasteners.locked
    def close(self):
        """Closes the contained jobboard, disallowing further use."""
        self._jobboard.close()

    @abc.abstractmethod
    def run(self, max_dispatches=None):
        """Continuously claims, runs, and consumes jobs (and repeat).

        :param max_dispatches: An upper bound on the number of jobs that will
                               be dispatched, if none or negative this implies
                               there is no limit to the number of jobs that
                               will be dispatched, otherwise if positive this
                               run method will return when that amount of jobs
                               has been dispatched (instead of running
                               forever and/or until stopped).
        """

    @abc.abstractmethod
    def _dispatch_job(self, job):
        """Dispatches a claimed job for work completion.

        Accepts a single (already claimed) job and causes it to be run in
        an engine. Returns a future object that represented the work to be
        completed sometime in the future. The future should return a single
        boolean from its result() method. This boolean determines whether the
        job will be consumed (true) or whether it should be abandoned (false).

        :param job: A job instance that has already been claimed by the
                    jobboard.
        """



# -*- coding: utf-8 -*-

#    Copyright (C) 2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
import time

from oslo_utils import timeutils

from taskflow.engines.action_engine import compiler as co
from taskflow import exceptions as exc
from taskflow.listeners import base
from taskflow import logging
from taskflow import states

STARTING_STATES = frozenset((states.RUNNING, states.REVERTING))
FINISHED_STATES = frozenset((base.FINISH_STATES + (states.REVERTED,)))
WATCH_STATES = frozenset(itertools.chain(FINISHED_STATES, STARTING_STATES,
                                         [states.PENDING]))

LOG = logging.getLogger(__name__)


# TODO(harlowja): get rid of this when we can just support python 3.x and use
# its print function directly instead of having to wrap it in a helper function
# due to how python 2.x print is a language built-in and not a function...
def _printer(message):
    print(message)


class DurationListener(base.Listener):
    """Listener that captures task duration.

    It records how long a task took to execute (or fail)
    to storage. It saves the duration in seconds as float value
    to task metadata with key ``'duration'``.
    """
    def __init__(self, engine):
        super(DurationListener, self).__init__(engine,
                                               task_listen_for=WATCH_STATES,
                                               flow_listen_for=WATCH_STATES)
        self._timers = {co.TASK: {}, co.FLOW: {}}

    def deregister(self):
        super(DurationListener, self).deregister()
        # There should be none that still exist at deregistering time, so log a
        # warning if there were any that somehow still got left behind...
        for item_type, timers in self._timers.items():
            leftover_timers = len(timers)
            if leftover_timers:
                LOG.warning("%s %s(s) did not enter %s states",
                            leftover_timers,
                            item_type, FINISHED_STATES)
            timers.clear()

    def _record_ending(self, timer, item_type, item_name, state):
        meta_update = {
            'duration': timer.elapsed(),
        }

        try:
            storage = self._engine.storage
            # Don't let storage failures throw exceptions in a listener method.
            if item_type == co.FLOW:
                storage.update_flow_metadata(meta_update)
            else:
                storage.update_atom_metadata(item_name, meta_update)
        except exc.StorageFailure:
            LOG.warning("Failure to store duration update %s for %s %s",
                        meta_update, item_type, item_name, exc_info=True)

    def _task_receiver(self, state, details):
        task_name = details['task_name']
        self._receiver(co.TASK, task_name, state)

    def _flow_receiver(self, state, details):
        flow_name = details['flow_name']
        self._receiver(co.FLOW, flow_name, state)

    def _receiver(self, item_type, item_name, state):
        if state == states.PENDING:
            self._timers[item_type].pop(item_name, None)
        elif state in STARTING_STATES:
            self._timers[item_type][item_name] = timeutils.StopWatch().start()
        elif state in FINISHED_STATES:
            timer = self._timers[item_type].pop(item_name, None)
            if timer is not None:
                timer.stop()
                self._record_ending(timer, item_type, item_name, state)


class PrintingDurationListener(DurationListener):
    """Listener that prints the duration as well as recording it."""

    def __init__(self, engine, printer=None):
        super(PrintingDurationListener, self).__init__(engine)
        if printer is None:
            self._printer = _printer
        else:
            self._printer = printer

    def _record_ending(self, timer, item_type, item_name, state):
        super(PrintingDurationListener, self)._record_ending(
            timer, item_type, item_name, state)
        self._printer("It took %s '%s' %0.2f seconds to"
                      " finish." % (item_type, item_name, timer.elapsed()))

    def _receiver(self, item_type, item_name, state):
        super(PrintingDurationListener, self)._receiver(item_type,
                                                        item_name, state)
        if state in STARTING_STATES:
            self._printer("'%s' %s started." % (item_name, item_type))


class EventTimeListener(base.Listener):
    """Listener that captures task, flow, and retry event timestamps.

    It records how when an event is received (using unix time) to
    storage. It saves the timestamps under keys (in atom or flow details
    metadata) of the format ``{event}-timestamp`` where ``event`` is the
    state/event name that has been received.

    This information can be later extracted/examined to derive durations...
    """

    def __init__(self, engine,
                 task_listen_for=base.DEFAULT_LISTEN_FOR,
                 flow_listen_for=base.DEFAULT_LISTEN_FOR,
                 retry_listen_for=base.DEFAULT_LISTEN_FOR):
        super(EventTimeListener, self).__init__(
            engine, task_listen_for=task_listen_for,
            flow_listen_for=flow_listen_for, retry_listen_for=retry_listen_for)

    def _record_atom_event(self, state, atom_name):
        meta_update = {'%s-timestamp' % state: time.time()}
        try:
            # Don't let storage failures throw exceptions in a listener method.
            self._engine.storage.update_atom_metadata(atom_name, meta_update)
        except exc.StorageFailure:
            LOG.warning("Failure to store timestamp %s for atom %s",
                        meta_update, atom_name, exc_info=True)

    def _flow_receiver(self, state, details):
        meta_update = {'%s-timestamp' % state: time.time()}
        try:
            # Don't let storage failures throw exceptions in a listener method.
            self._engine.storage.update_flow_metadata(meta_update)
        except exc.StorageFailure:
            LOG.warning("Failure to store timestamp %s for flow %s",
                        meta_update, details['flow_name'], exc_info=True)

    def _task_receiver(self, state, details):
        self._record_atom_event(state, details['task_name'])

    def _retry_receiver(self, state, details):
        self._record_atom_event(state, details['retry_name'])



# -*- coding: utf-8 -*-

#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import copy
import io
import os
import sys
import traceback

from oslo_utils import reflection

from taskflow import exceptions as exc
from taskflow.utils import iter_utils
from taskflow.utils import schema_utils as su


def _copy_exc_info(exc_info):
    if exc_info is None:
        return None
    exc_type, exc_value, tb = exc_info
    # NOTE(imelnikov): there is no need to copy the exception type, and
    # a shallow copy of the value is fine and we can't copy the traceback since
    # it contains reference to the internal stack frames...
    return (exc_type, copy.copy(exc_value), tb)


def _are_equal_exc_info_tuples(ei1, ei2):
    if ei1 == ei2:
        return True
    if ei1 is None or ei2 is None:
        return False  # if both are None, we returned True above

    # NOTE(imelnikov): we can't compare exceptions with '=='
    # because we want exc_info be equal to it's copy made with
    # copy_exc_info above.
    if ei1[0] is not ei2[0]:
        return False
    # NOTE(dhellmann): The flake8/pep8 error E721 does not apply here
    # because we want the types to be exactly the same, not just have
    # one be inherited from the other.
    if not all((type(ei1[1]) == type(ei2[1]),  # noqa: E721
                str(ei1[1]) == str(ei2[1]),
                repr(ei1[1]) == repr(ei2[1]))):
        return False
    if ei1[2] == ei2[2]:
        return True
    tb1 = traceback.format_tb(ei1[2])
    tb2 = traceback.format_tb(ei2[2])
    return tb1 == tb2


class Failure():
    """An immutable object that represents failure.

    Failure objects encapsulate exception information so that they can be
    re-used later to re-raise, inspect, examine, log, print, serialize,
    deserialize...

    One example where they are depended upon is in the WBE engine. When a
    remote worker throws an exception, the WBE based engine will receive that
    exception and desire to reraise it to the user/caller of the WBE based
    engine for appropriate handling (this matches the behavior of non-remote
    engines). To accomplish this a failure object (or a
    :py:meth:`~.Failure.to_dict` form) would be sent over the WBE channel
    and the WBE based engine would deserialize it and use this objects
    :meth:`.reraise` method to cause an exception that contains
    similar/equivalent information as the original exception to be reraised,
    allowing the user (or the WBE engine itself) to then handle the worker
    failure/exception as they desire.

    For those who are curious, here are a few reasons why the original
    exception itself *may* not be reraised and instead a reraised wrapped
    failure exception object will be instead. These explanations are *only*
    applicable when a failure object is serialized and deserialized (when it is
    retained inside the python process that the exception was created in the
    the original exception can be reraised correctly without issue).

    * Traceback objects are not serializable/recreatable, since they contain
      references to stack frames at the location where the exception was
      raised. When a failure object is serialized and sent across a channel
      and recreated it is *not* possible to restore the original traceback and
      originating stack frames.
    * The original exception *type* can not be guaranteed to be found, workers
      can run code that is not accessible/available when the failure is being
      deserialized. Even if it was possible to use pickle safely it would not
      be possible to find the originating exception or associated code in this
      situation.
    * The original exception *type* can not be guaranteed to be constructed in
      a *correct* manner. At the time of failure object creation the exception
      has already been created and the failure object can not assume it has
      knowledge (or the ability) to recreate the original type of the captured
      exception (this is especially hard if the original exception was created
      via a complex process via some custom exception constructor).
    * The original exception *type* can not be guaranteed to be constructed in
      a *safe* manner. Importing *foreign* exception types dynamically can be
      problematic when not done correctly and in a safe manner; since failure
      objects can capture any exception it would be *unsafe* to try to import
      those exception types namespaces and modules on the receiver side
      dynamically (this would create similar issues as the ``pickle`` module in
      python has where foreign modules can be imported, causing those modules
      to have code ran when this happens, and this can cause issues and
      side-effects that the receiver would not have intended to have caused).

    TODO(harlowja): use parts of :pybug:`17911` and the
    backport at https://pypi.org/project/traceback2/ to (hopefully)
    simplify the methods and contents of this object...
    """
    DICT_VERSION = 1

    BASE_EXCEPTIONS = ('BaseException', 'Exception')
    """
    Root exceptions of all other python exceptions.

    See: https://docs.python.org/2/library/exceptions.html
    """

    #: Expected failure schema (in json schema format).
    SCHEMA = {
        "$ref": "#/definitions/cause",
        "definitions": {
            "cause": {
                "type": "object",
                'properties': {
                    'version': {
                        "type": "integer",
                        "minimum": 0,
                    },
                    'exc_args': {
                        "type": "array",
                        "minItems": 0,
                    },
                    'exception_str': {
                        "type": "string",
                    },
                    'traceback_str': {
                        "type": "string",
                    },
                    'exc_type_names': {
                        "type": "array",
                        "items": {
                            "type": "string",
                        },
                        "minItems": 1,
                    },
                    'causes': {
                        "type": "array",
                        "items": {
                            "$ref": "#/definitions/cause",
                        },
                    }
                },
                "required": [
                    "exception_str",
                    'traceback_str',
                    'exc_type_names',
                ],
                "additionalProperties": True,
            },
        },
    }

    def __init__(self, exc_info=None, **kwargs):
        if not kwargs:
            if exc_info is None:
                exc_info = sys.exc_info()
            else:
                # This should always be the (type, value, traceback) tuple,
                # either from a prior sys.exc_info() call or from some other
                # creation...
                if len(exc_info) != 3:
                    raise ValueError("Provided 'exc_info' must contain three"
                                     " elements")
            self._exc_info = exc_info
            self._exc_args = tuple(getattr(exc_info[1], 'args', []))
            self._exc_type_names = tuple(
                reflection.get_all_class_names(exc_info[0], up_to=Exception))
            if not self._exc_type_names:
                raise TypeError("Invalid exception type '%s' (%s)"
                                % (exc_info[0], type(exc_info[0])))
            self._exception_str = str(self._exc_info[1])
            self._traceback_str = ''.join(
                traceback.format_tb(self._exc_info[2]))
            self._causes = kwargs.pop('causes', None)
        else:
            self._causes = kwargs.pop('causes', None)
            self._exc_info = exc_info
            self._exc_args = tuple(kwargs.pop('exc_args', []))
            self._exception_str = kwargs.pop('exception_str')
            self._exc_type_names = tuple(kwargs.pop('exc_type_names', []))
            self._traceback_str = kwargs.pop('traceback_str', None)
            if kwargs:
                raise TypeError(
                    'Failure.__init__ got unexpected keyword argument(s): %s'
                    % ', '.join(kwargs.keys()))

    @classmethod
    def from_exception(cls, exception):
        """Creates a failure object from a exception instance."""
        exc_info = (
            type(exception),
            exception,
            getattr(exception, '__traceback__', None)
        )
        return cls(exc_info=exc_info)

    @classmethod
    def validate(cls, data):
        """Validate input data matches expected failure ``dict`` format."""
        try:
            su.schema_validate(data, cls.SCHEMA)
        except su.ValidationError as e:
            raise exc.InvalidFormat("Failure data not of the"
                                    " expected format: %s" % (e.message), e)
        else:
            # Ensure that all 'exc_type_names' originate from one of
            # BASE_EXCEPTIONS, because those are the root exceptions that
            # python mandates/provides and anything else is invalid...
            causes = collections.deque([data])
            while causes:
                cause = causes.popleft()
                root_exc_type = cause['exc_type_names'][-1]
                if root_exc_type not in cls.BASE_EXCEPTIONS:
                    raise exc.InvalidFormat(
                        "Failure data 'exc_type_names' must"
                        " have an initial exception type that is one"
                        " of %s types: '%s' is not one of those"
                        " types" % (cls.BASE_EXCEPTIONS, root_exc_type))
                sub_causes = cause.get('causes')
                if sub_causes:
                    causes.extend(sub_causes)

    def _matches(self, other):
        if self is other:
            return True
        return (self._exc_type_names == other._exc_type_names
                and self.exception_args == other.exception_args
                and self.exception_str == other.exception_str
                and self.traceback_str == other.traceback_str
                and self.causes == other.causes)

    def matches(self, other):
        """Checks if another object is equivalent to this object.

        :returns: checks if another object is equivalent to this object
        :rtype: boolean
        """
        if not isinstance(other, Failure):
            return False
        if self.exc_info is None or other.exc_info is None:
            return self._matches(other)
        else:
            return self == other

    def __eq__(self, other):
        if not isinstance(other, Failure):
            return NotImplemented
        return (self._matches(other) and
                _are_equal_exc_info_tuples(self.exc_info, other.exc_info))

    def __ne__(self, other):
        return not (self == other)

    # NOTE(imelnikov): obj.__hash__() should return same values for equal
    # objects, so we should redefine __hash__. Failure equality semantics
    # is a bit complicated, so for now we just mark Failure objects as
    # unhashable. See python docs on object.__hash__  for more info:
    # http://docs.python.org/2/reference/datamodel.html#object.__hash__
    __hash__ = None

    @property
    def exception(self):
        """Exception value, or none if exception value is not present.

        Exception value may be lost during serialization.
        """
        if self._exc_info:
            return self._exc_info[1]
        else:
            return None

    @property
    def exception_str(self):
        """String representation of exception."""
        return self._exception_str

    @property
    def exception_args(self):
        """Tuple of arguments given to the exception constructor."""
        return self._exc_args

    @property
    def exc_info(self):
        """Exception info tuple or none.

        See: https://docs.python.org/2/library/sys.html#sys.exc_info for what
             the contents of this tuple are (if none, then no contents can
             be examined).
        """
        return self._exc_info

    @property
    def traceback_str(self):
        """Exception traceback as string."""
        return self._traceback_str

    @staticmethod
    def reraise_if_any(failures):
        """Re-raise exceptions if argument is not empty.

        If argument is empty list/tuple/iterator, this method returns
        None. If argument is converted into a list with a
        single ``Failure`` object in it, that failure is reraised. Else, a
        :class:`~taskflow.exceptions.WrappedFailure` exception
        is raised with the failure list as causes.
        """
        if not isinstance(failures, (list, tuple)):
            # Convert generators/other into a list...
            failures = list(failures)
        if len(failures) == 1:
            failures[0].reraise()
        elif len(failures) > 1:
            raise exc.WrappedFailure(failures)

    def reraise(self):
        """Re-raise captured exception."""
        if self._exc_info:
            tp, value, tb = self._exc_info
            if value is None:
                value = tp()
            if value.__traceback__ is not tb:
                raise value.with_traceback(tb)
            raise value
        else:
            raise exc.WrappedFailure([self])

    def check(self, *exc_classes):
        """Check if any of ``exc_classes`` caused the failure.

        Arguments of this method can be exception types or type
        names (stings). If captured exception is instance of
        exception of given type, the corresponding argument is
        returned. Else, None is returned.
        """
        for cls in exc_classes:
            if isinstance(cls, type):
                err = reflection.get_class_name(cls)
            else:
                err = cls
            if err in self._exc_type_names:
                return cls
        return None

    @classmethod
    def _extract_causes_iter(cls, exc_val):
        seen = [exc_val]
        causes = [exc_val]
        while causes:
            exc_val = causes.pop()
            if exc_val is None:
                continue
            # See: https://www.python.org/dev/peps/pep-3134/ for why/what
            # these are...
            #
            # '__cause__' attribute for explicitly chained exceptions
            # '__context__' attribute for implicitly chained exceptions
            # '__traceback__' attribute for the traceback
            #
            # See: https://www.python.org/dev/peps/pep-0415/ for why/what
            # the '__suppress_context__' is/means/implies...
            suppress_context = getattr(exc_val,
                                       '__suppress_context__', False)
            if suppress_context:
                attr_lookups = ['__cause__']
            else:
                attr_lookups = ['__cause__', '__context__']
            nested_exc_val = None
            for attr_name in attr_lookups:
                attr_val = getattr(exc_val, attr_name, None)
                if attr_val is None:
                    continue
                if attr_val not in seen:
                    nested_exc_val = attr_val
                    break
            if nested_exc_val is not None:
                exc_info = (
                    type(nested_exc_val),
                    nested_exc_val,
                    getattr(nested_exc_val, '__traceback__', None),
                )
                seen.append(nested_exc_val)
                causes.append(nested_exc_val)
                yield cls(exc_info=exc_info)

    @property
    def causes(self):
        """Tuple of all *inner* failure *causes* of this failure.

        NOTE(harlowja): Does **not** include the current failure (only
        returns connected causes of this failure, if any). This property
        is really only useful on 3.x or newer versions of python as older
        versions do **not** have associated causes (the tuple will **always**
        be empty on 2.x versions of python).

        Refer to :pep:`3134` and :pep:`409` and :pep:`415` for what
        this is examining to find failure causes.
        """
        if self._causes is not None:
            return self._causes
        else:
            self._causes = tuple(self._extract_causes_iter(self.exception))
            return self._causes

    def __str__(self):
        return self.pformat()

    def pformat(self, traceback=False):
        """Pretty formats the failure object into a string."""
        buf = io.StringIO()
        if not self._exc_type_names:
            buf.write('Failure: %s' % (self._exception_str))
        else:
            buf.write('Failure: %s: %s' % (self._exc_type_names[0],
                                           self._exception_str))
        if traceback:
            if self._traceback_str is not None:
                traceback_str = self._traceback_str.rstrip()
            else:
                traceback_str = None
            if traceback_str:
                buf.write(os.linesep)
                buf.write('Traceback (most recent call last):')
                buf.write(os.linesep)
                buf.write(traceback_str)
            else:
                buf.write(os.linesep)
                buf.write('Traceback not available.')
        return buf.getvalue()

    def __iter__(self):
        """Iterate over exception type names."""
        for et in self._exc_type_names:
            yield et

    def __getstate__(self):
        dct = self.to_dict()
        if self._exc_info:
            # Avoids 'TypeError: can't pickle traceback objects'
            dct['exc_info'] = self._exc_info[0:2]
        return dct

    def __setstate__(self, dct):
        self._exception_str = dct['exception_str']
        if 'exc_args' in dct:
            self._exc_args = tuple(dct['exc_args'])
        else:
            # Guess we got an older version somehow, before this
            # was added, so at that point just set to an empty tuple...
            self._exc_args = ()
        self._traceback_str = dct['traceback_str']
        self._exc_type_names = dct['exc_type_names']
        if 'exc_info' in dct:
            # Tracebacks can't be serialized/deserialized, but since we
            # provide a traceback string (and more) this should be
            # acceptable...
            #
            # TODO(harlowja): in the future we could do something like
            # what the twisted people have done, see for example
            # twisted-13.0.0/twisted/python/failure.py#L89 for how they
            # created a fake traceback object...
            self._exc_info = tuple(iter_utils.fill(dct['exc_info'], 3))
        else:
            self._exc_info = None
        causes = dct.get('causes')
        if causes is not None:
            causes = tuple(self.from_dict(d) for d in causes)
        self._causes = causes

    @classmethod
    def from_dict(cls, data):
        """Converts this from a dictionary to a object."""
        data = dict(data)
        version = data.pop('version', None)
        if version != cls.DICT_VERSION:
            raise ValueError('Invalid dict version of failure object: %r'
                             % version)
        causes = data.get('causes')
        if causes is not None:
            data['causes'] = tuple(cls.from_dict(d) for d in causes)
        return cls(**data)

    def to_dict(self, include_args=True):
        """Converts this object to a dictionary.

        :param include_args: boolean indicating whether to include the
                             exception args in the output.
        """
        return {
            'exception_str': self.exception_str,
            'traceback_str': self.traceback_str,
            'exc_type_names': list(self),
            'version': self.DICT_VERSION,
            'exc_args': self.exception_args if include_args else tuple(),
            'causes': [f.to_dict() for f in self.causes],
        }

    def copy(self):
        """Copies this object."""
        return Failure(exc_info=_copy_exc_info(self.exc_info),
                       exception_str=self.exception_str,
                       traceback_str=self.traceback_str,
                       exc_args=self.exception_args,
                       exc_type_names=self._exc_type_names[:],
                       causes=self._causes)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Database setup and migration commands."""

import os

from alembic import command
from alembic import config


def _make_alembic_config():
    path = os.path.join(os.path.dirname(__file__), 'alembic', 'alembic.ini')
    return config.Config(path)


def db_sync(connection, revision='head'):
    cfg = _make_alembic_config()
    cfg.attributes['connection'] = connection
    command.upgrade(cfg, revision)
